import json
import logging
import math
import os
import pdb
import random
import re
import sys
import time
import traceback
from collections import defaultdict
from typing import Dict, List, Optional, Sequence

import numpy as np
import torch
import transformers
from transformers.trainer_pt_utils import LabelSmoother

from .dataset_base import BaseDataset

IGNORE_TOKEN_ID = LabelSmoother.ignore_index


class DeepSeekDataset(BaseDataset):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )

        self.default_system_message = "You are a helpful AI assistant."
        self.default_system_message = None

        self.ret = defaultdict(dict)
        self.is_cat = True

        if self.cross_dataset_joint:
            for i in range(2):
                self.maybe_init_ret(f"default_{i}")

    def maybe_init_ret(self, source, force=False):
        if source not in self.ret or force:
            self.ret[source] = {}

            self.ret[source]["tokens"] = []
            self.ret[source]["labels"] = []
            self.ret[source]["actual_seq_len"] = []

            if self.create_position_ids:
                self.ret[source]["position_ids"] = []

            if self.create_attention_mask:
                self.ret[source]["attention_mask"] = []

            if self.create_attention_mask_2d:
                self.ret[source]["attention_mask_2d"] = torch.tril(
                    torch.ones(
                        (1, self.max_padding_length, self.max_padding_length), dtype=torch.bool
                    )
                )
        return len(self.ret[source]["tokens"]) == 0

    def get_max_min_ret_length(self):
        max_ret_lengh = 0
        min_ret_lengh = self.max_padding_length + 1

        max_ret_key = None
        min_ret_key = None

        for k, v in self.ret.items():
            cur_length = len(v["tokens"])

            if cur_length > max_ret_lengh:
                max_ret_lengh = cur_length
                max_ret_key = k

            if cur_length < min_ret_lengh:
                min_ret_lengh = cur_length
                min_ret_key = k

        return max_ret_lengh, max_ret_key, min_ret_lengh, min_ret_key

    def add_ret(self, ret, source):
        cur_length = len(ret["input_ids"])
        cur_image_length = len(ret["images"])
        cur_audio_length = len(ret["audios"])

        all_length = len(self.ret[source]["tokens"])

        if "images" in self.ret[source]:
            all_image_length = len(self.ret[source]["images"])
        else:
            all_image_length = 0

        if cur_image_length > 0:
            if all_image_length > 0:
                self.ret[source]["images"] = torch.cat(
                    [self.ret[source]["images"], ret["images"]], dim=0
                )
                ret["image_indices"][1, :, :] += all_length
                self.ret[source]["image_indices"] = torch.cat(
                    [self.ret[source]["image_indices"], ret["image_indices"]], dim=1
                )
            else:
                self.ret[source]["images"] = ret["images"]
                self.ret[source]["image_indices"] = ret["image_indices"]

        if "audios" in self.ret[source]:
            all_audio_length = len(self.ret[source]["audios"])
        else:
            all_audio_length = 0

        if cur_audio_length > 0:
            if all_audio_length > 0:
                # self.ret[source]["audios"] = torch.cat(
                #     [self.ret[source]["audios"], ret["audios"]], dim=0
                # )
                # ret["audio_indices"][1, :, :] += all_length
                # self.ret[source]["audio_indices"] = torch.cat(
                #     [self.ret[source]["audio_indices"], ret["audio_indices"]], dim=1
                # )
                self.ret[source]["audios"].extend(ret["audios"])
                for audio_indice in ret["audio_indices"]:
                    audio_indice[1, :, :] += all_length
                self.ret[source]["audio_indices"].extend(ret["audio_indices"])
            else:
                self.ret[source]["audios"] = ret["audios"]
                self.ret[source]["audio_indices"] = ret["audio_indices"]

            # print(self.ret[source]["audios"])

        if self.create_attention_mask:
            self.ret[source]["attention_mask"] += ret["attention_mask"]

        if self.create_attention_mask_2d:
            self.ret[source]["attention_mask_2d"][:, all_length:, :all_length] = 0

        if self.create_position_ids:
            self.ret[source]["position_ids"] += list(range(cur_length))

        self.ret[source]["tokens"] += ret["input_ids"]
        self.ret[source]["labels"] += ret["labels"]
        self.ret[source]["actual_seq_len"] += [all_length + cur_length]

    def process_ret(self, to_ret):
        if "tokens" in to_ret and len(to_ret["tokens"]) > 0:
            pass
        else:
            return to_ret

        if self.create_position_ids:
            if self.reset_position_ids:
                pass
            else:
                to_ret["position_ids"] = list(range(len(to_ret["tokens"])))

        if self.create_attention_mask_2d:
            if self.reset_attention_mask:
                pass
            else:
                to_ret["attention_mask_2d"] = torch.tril(
                    torch.ones(
                        (1, self.max_padding_length, self.max_padding_length), dtype=torch.bool
                    )
                )

        if self.shift_token:
            to_ret["tokens"] = to_ret["tokens"][:-1]
            to_ret["labels"] = to_ret["labels"][1:]
            to_ret["actual_seq_len"][-1] -= 1
            if self.create_position_ids:
                to_ret["position_ids"] = to_ret["position_ids"][:-1]
            if self.create_attention_mask:
                to_ret["attention_mask"] = to_ret["attention_mask"][:-1]

            if self.create_attention_mask_2d:
                to_ret["attention_mask_2d"][:, :, -1] = 0
                to_ret["attention_mask_2d"][:, -1, :] = 0

        assert len(to_ret["tokens"]) == len(
            to_ret["labels"]
        ), f"{len(to_ret['tokens'])} {len(to_ret['labels'])}"
        if not self.variable_length and self.max_padding_length > len(to_ret["tokens"]):
            to_ret["tokens"] += [self.tokenizer.pad_token_id] * (
                self.max_padding_length - len(to_ret["tokens"])
            )
            to_ret["labels"] += [IGNORE_TOKEN_ID] * (
                self.max_padding_length - len(to_ret["labels"])
            )
            to_ret["actual_seq_len"][-1] = self.max_padding_length
            if self.create_position_ids:
                # to_ret["position_ids"] += to_ret["position_ids"][-1:] * (
                #     self.max_padding_length - len(to_ret["position_ids"])
                # )
                to_ret["position_ids"] += list(
                    range(to_ret["position_ids"][-1] + 1, self.max_padding_length)
                )
            if self.create_attention_mask:
                to_ret["attention_mask"] += [0] * (
                    self.max_padding_length - len(to_ret["attention_mask"])
                )

        to_ret["tokens"] = to_ret["tokens"][: self.max_padding_length]
        to_ret["labels"] = to_ret["labels"][: self.max_padding_length]
        to_ret["actual_seq_len"][-1] = self.max_padding_length
        if self.create_position_ids:
            to_ret["position_ids"] = to_ret["position_ids"][: self.max_padding_length]
        if self.create_attention_mask:
            to_ret["attention_mask"] = to_ret["attention_mask"][: self.max_padding_length]

        to_ret["tokens"] = torch.tensor(to_ret["tokens"], dtype=torch.int64)
        to_ret["labels"] = torch.tensor(to_ret["labels"], dtype=torch.int64)
        to_ret["actual_seq_len"] = torch.tensor(to_ret["actual_seq_len"], dtype=torch.int64)
        if self.create_position_ids:
            to_ret["position_ids"] = torch.tensor(to_ret["position_ids"], dtype=torch.int64)
        if self.create_attention_mask:
            to_ret["attention_mask"] = torch.tensor(to_ret["attention_mask"], dtype=torch.int64)

        if self.create_attention_mask_2d:
            attention_mask_2d = to_ret.pop("attention_mask_2d")
            attention_mask_2d = attention_mask_2d.masked_fill(
                (to_ret["attention_mask"] < 0.5).view(1, 1, self.max_padding_length), value=0
            )
            attention_mask_2d = attention_mask_2d < 0.5

            to_ret["attention_mask"] = attention_mask_2d

        if self.create_loss_mask:
            loss_mask = torch.where(to_ret["labels"] == IGNORE_TOKEN_ID, 0, 1)
            to_ret["loss_mask"] = loss_mask.to(torch.float32)

        if not self.reset_position_ids and not self.reset_attention_mask:
            to_ret.pop("actual_seq_len")

        to_ret["input_ids"] = to_ret["tokens"]

        # print("to_ret[tokens]", to_ret["tokens"])
        # print("to_ret[labels]", to_ret["labels"])

        return to_ret

    def is_skip(self):
        if self.processed_samples < self.skip_samples:
            if self.processed_samples % 1e3 == 0:
                print(
                    f"processed_samples {self.processed_samples} skip_samples {self.skip_samples}"
                )
            return True

    def show_statistic(self):
        log_interval = 10000
        if self.max_padding_length >= 2**17:
            log_interval = 500
        if self.max_padding_length >= 2**20:
            log_interval = 100

        if self.unjoint_samples % log_interval == 0:
            print(
                f"processed_samples {self.processed_samples} unjoint_samples {self.unjoint_samples} joint_samples {self.joint_samples} {[len(v['tokens']) for _, v in self.ret.items()]}",
                flush=True,
            )

        return False

    def __getitem__(self, index):

        self.processor["audio"].load_model()

        while True:
            # if True:
            try:
                self.processed_samples += 1
                if self.is_skip():
                    return {}

                sample = self.raw_data[index]

                if self.cross_dataset_joint:
                    is_empty = False
                    (
                        max_ret_lengh,
                        max_ret_key,
                        min_ret_lengh,
                        min_ret_key,
                    ) = self.get_max_min_ret_length()
                else:
                    source = sample["source"]
                    is_empty = self.maybe_init_ret(source)

                    max_ret_lengh = min_ret_lengh = len(self.ret[source]["tokens"])
                    max_ret_key = min_ret_key = source

                is_begin = is_empty or self.reset_position_ids or self.reset_attention_mask

                ret = preprocess(
                    sample,
                    self.tokenizer,
                    self.image_token_length,
                    default_system_message=self.default_system_message,
                    processor=self.processor,
                    is_begin=is_begin,
                    max_num_frame=self.max_num_frame,
                    max_fps=self.max_fps,
                )

                if ret is None:
                    return {}

                cur_length = len(ret["input_ids"])

                if cur_length > self.max_padding_length:
                    return {}

                self.unjoint_samples += 1

                if not self.dataset_joint:
                    to_ret = self.ret.pop(max_ret_key)

                    self.maybe_init_ret(max_ret_key, force=True)
                    self.add_ret(ret, max_ret_key)

                elif min_ret_lengh + cur_length > self.max_padding_length:
                    to_ret = self.ret.pop(max_ret_key)
                    self.joint_samples += 1

                    self.maybe_init_ret(max_ret_key, force=True)
                    self.add_ret(ret, max_ret_key)

                else:
                    to_ret = {}
                    self.add_ret(ret, min_ret_key)

                to_ret = self.process_ret(to_ret)

                self.show_statistic()
                return to_ret

            except Exception as error:
                try:
                    with open(os.path.join(self.output_dir, "data_error.log"), "a") as f:
                        print("-" * 100, file=f)
                        print(traceback.format_exc(), file=f)
                        print(self.raw_data[index], file=f)
                except Exception as error:
                    print(error)
                return {}


def preprocess(
    sample,
    tokenizer: transformers.PreTrainedTokenizer,
    image_token_length: int,
    default_system_message: str = "You are a helpful assistant.",
    processor=None,
    is_begin: bool = True,
    max_num_frame: int = 8,
    max_fps: int = 1,
) -> Dict:

    # <|im_start|>system
    # You are Qwen, created by Alibaba Cloud. You are a helpful assistant.<|im_end|>
    # <|im_start|>user
    # Hello, how are you?<|im_end|>
    # <|im_start|>assistantI'm doing great. How can I help you today?<|im_end|>
    # <|im_start|>user
    # I'd like to show off how chat templating works!<|im_end|>

    from ..constants import (
        IMG_START_TOKEN,
        IMG_END_TOKEN,
        IMG_CONTEXT_TOKEN,
        VID_START_TOKEN,
        VID_END_TOKEN,
        VID_CONTEXT_TOKEN,
        PATCH_START_TOKEN,
        PATCH_END_TOKEN,
        PATCH_CONTEXT_TOKEN,
        AUD_START_TOKEN,
        AUD_END_TOKEN,
        IMG_TAG_TOKEN,
        VID_TAG_TOKEN,
        AUD_TAG_TOKEN,
        AUD_CONTEXT_TOKEN,
    )

    human_roles = ["user", "human"]
    gpt_roles = ["assistant", "gpt"]
    system_roles = ["system"]

    IMG_CONTEXT_ID = tokenizer(IMG_CONTEXT_TOKEN, add_special_tokens=False).input_ids
    IMG_START_ID = tokenizer(IMG_START_TOKEN, add_special_tokens=False).input_ids
    IMG_END_ID = tokenizer(IMG_END_TOKEN, add_special_tokens=False).input_ids

    VID_CONTEXT_ID = tokenizer(VID_CONTEXT_TOKEN, add_special_tokens=False).input_ids
    VID_START_ID = tokenizer(VID_START_TOKEN, add_special_tokens=False).input_ids
    VID_END_ID = tokenizer(VID_END_TOKEN, add_special_tokens=False).input_ids

    PATCH_CONTEXT_ID = tokenizer(PATCH_CONTEXT_TOKEN, add_special_tokens=False).input_ids
    PATCH_START_ID = tokenizer(PATCH_START_TOKEN, add_special_tokens=False).input_ids
    PATCH_END_ID = tokenizer(PATCH_END_TOKEN, add_special_tokens=False).input_ids

    AUD_CONTEXT_ID = tokenizer(AUD_CONTEXT_TOKEN, add_special_tokens=False).input_ids
    AUD_START_ID = tokenizer(AUD_START_TOKEN, add_special_tokens=False).input_ids
    AUD_END_ID = tokenizer(AUD_END_TOKEN, add_special_tokens=False).input_ids

    IMG_TAG_ID = tokenizer(IMG_TAG_TOKEN, add_special_tokens=False).input_ids
    VID_TAG_ID = tokenizer(VID_TAG_TOKEN, add_special_tokens=False).input_ids
    AUD_TAG_ID = tokenizer(AUD_TAG_TOKEN, add_special_tokens=False).input_ids

    assert len(IMG_CONTEXT_ID) == 1
    assert len(IMG_START_ID) == 1
    assert len(IMG_END_ID) == 1

    assert len(VID_CONTEXT_ID) == 1
    assert len(VID_START_ID) == 1
    assert len(VID_END_ID) == 1

    assert len(PATCH_CONTEXT_ID) == 1
    assert len(PATCH_START_ID) == 1
    assert len(PATCH_END_ID) == 1

    IMG_CONTEXT_ID = IMG_CONTEXT_ID[0]
    IMG_START_ID = IMG_START_ID[0]
    IMG_END_ID = IMG_END_ID[0]

    VID_CONTEXT_ID = VID_CONTEXT_ID[0]
    VID_START_ID = VID_START_ID[0]
    VID_END_ID = VID_END_ID[0]

    PATCH_CONTEXT_ID = PATCH_CONTEXT_ID[0]
    PATCH_START_ID = PATCH_START_ID[0]
    PATCH_END_ID = PATCH_END_ID[0]

    AUD_CONTEXT_ID = AUD_CONTEXT_ID[0]
    AUD_START_ID = AUD_START_ID[0]
    AUD_END_ID = AUD_END_ID[0]

    IMG_TAG_ID = IMG_TAG_ID[0]
    VID_TAG_ID = VID_TAG_ID[0]
    AUD_TAG_ID = AUD_TAG_ID[0]

    BOS_ID = tokenizer.bos_token_id
    EOS_ID = tokenizer.eos_token_id

    IM_START = "<｜begin▁of▁sentence｜>"
    IM_END = "<｜end▁of▁sentence｜>"
    USER = "<｜User｜>"
    ASSISTANT = "<｜Assistant｜>"

    nl_tokens = tokenizer("\n", add_special_tokens=False).input_ids
    IM_START_IDS = tokenizer(IM_START, add_special_tokens=False).input_ids
    IM_END_IDS = tokenizer(IM_END, add_special_tokens=False).input_ids
    USER_IDS = tokenizer(USER, add_special_tokens=False).input_ids
    ASSISTANT_IDS = tokenizer(ASSISTANT, add_special_tokens=False).input_ids

    assert len(USER_IDS) == 1, USER_IDS
    assert len(ASSISTANT_IDS) == 1, ASSISTANT_IDS
    assert len(IM_END_IDS) == 1, IM_END_IDS
    assert len(IM_START_IDS) == 1, IM_START_IDS

    input_ids, targets = [], []
    images = []
    image_indices = []
    audios = []
    audio_indices = []

    messages = []
    if "conversations" in sample:
        messages = sample["conversations"]
    if len(messages) == 0 and "messages" in sample:
        messages = sample["messages"]

    # ----------------------------------------------------------------
    # add text to TTS
    if True:
        add_text = None
        # add_audio = None
        for j, sentence in enumerate(messages):
            content = sentence["content"]
            role = sentence["role"]
            if role == "user":
                if "Convert the text to speech." in content:
                    add_text = content.replace("Convert the text to speech.\n", "")
                    add_text = add_text.strip()

                # if "Convert the speech to text." in content:
                #     add_audio = sample["audios"][-1]

            if role == "assistant" and add_text is not None:
                sentence["content"] = add_text + content

            # if role == "assistant" and add_audio is not None:
            #     sentence["content"] = content + "\n<audio>"
            #     sample["audios"].append(add_audio)

    # ----------------------------------------------------------------
    # system
    has_system = False
    if is_begin:
        if messages[0]["role"] == "system":
            has_system = True
        else:
            has_system = False

        if (
            not has_system
            and default_system_message is not None
            and len(default_system_message) > 0
        ):
            messages = [{"role": "system", "content": default_system_message}] + messages
            has_system = True

    # ----------------------------------------------------------------
    # audio
    if has_audio(sample) and processor["audio"].is_discrete:
        unused_audio_idxs = list(range(len(sample["audios"])))
        audio_tokens_list = [
            processor["audio"].process_audios(x, is_discrete=True) for x in sample["audios"]
        ]
        audio_tokens_list = ["".join(f"<|audio_{i}|>" for i in x) for x in audio_tokens_list]

        audio_idx = 0
        for j, sentence in enumerate(messages):
            content = sentence["content"]
            role = sentence["role"]
            # whether apply discrete tokenize to this role
            if processor["audio"].apply_to_role(role, is_discrete=True):
                while AUD_TAG_TOKEN in content:
                    content = content.replace(
                        AUD_TAG_TOKEN,
                        f"{AUD_START_TOKEN}{audio_tokens_list[audio_idx]}{AUD_END_TOKEN}",
                        1,
                    )
                    unused_audio_idxs.remove(audio_idx)
                    audio_idx += 1
            else:
                audio_idx += content.count(AUD_TAG_TOKEN)

            sentence["content"] = content

    # ----------------------------------------------------------------
    input_ids += IM_START_IDS
    targets = [IGNORE_TOKEN_ID] * len(IM_START_IDS)
    # text
    for j, sentence in enumerate(messages):
        role = sentence["role"]
        content = sentence["content"]

        if role in human_roles:
            _input_id = USER_IDS + tokenizer(content, add_special_tokens=False).input_ids
            _target = [IGNORE_TOKEN_ID] * len(_input_id)

        elif role in gpt_roles:
            content_input_id = tokenizer(content, add_special_tokens=False).input_ids
            if processor["audio"].audio_tokenizer is not None:
                content_input_id = text_audio_interval(
                    content_input_id,
                    AUD_START_ID,
                    AUD_END_ID,
                    processor["audio"].text_audio_interval_ratio,
                )

            _input_id = ASSISTANT_IDS + content_input_id + IM_END_IDS
            _target = [IGNORE_TOKEN_ID] * len(ASSISTANT_IDS) + content_input_id + IM_END_IDS

        elif role in system_roles:
            if "神农" in content and "腾讯" in content:
                continue
            _input_id = tokenizer(content, add_special_tokens=False).input_ids
            _target = [IGNORE_TOKEN_ID] * len(_input_id)

        else:
            raise NotImplementedError

        # print(f"_input_id {_input_id}")
        input_ids += _input_id
        targets += _target

    # ----------------------------------------------------------------
    # image
    if has_image(sample):
        img_positions = [i for i, x in enumerate(input_ids) if x == IMG_TAG_ID]
        assert len(img_positions) == len(sample["images"]), sample

        new_input_ids = []
        new_targets = []
        st = 0
        for img_idx, img_pos in enumerate(img_positions):
            image_patches, (best_width, best_height) = processor[
                "image"
            ].process_images_with_subpatch(sample["images"][img_idx])
            images.append(image_patches)

            new_input_ids += input_ids[st:img_pos]
            new_targets += targets[st:img_pos]

            new_input_ids += [IMG_START_ID]
            new_targets += [IGNORE_TOKEN_ID]

            image_indice_b = torch.zeros(
                1, image_token_length, dtype=torch.int64
            )  # This will change in collate_fn
            image_indice_s = (
                torch.arange(len(new_input_ids), len(new_input_ids) + image_token_length)
                .unsqueeze(0)
                .repeat(1, 1)
            )
            image_indice_b_s = torch.stack(
                [image_indice_b, image_indice_s], dim=0
            )  # 2, num_image, image_length
            image_indices.append(image_indice_b_s)

            new_input_ids += [IMG_CONTEXT_ID] * image_token_length
            new_targets += [IGNORE_TOKEN_ID] * image_token_length

            new_input_ids += [IMG_END_ID]
            new_targets += [IGNORE_TOKEN_ID]

            if len(image_patches) > 1:
                for i in range(0, best_height, processor["image"].patch_size):
                    new_input_ids += nl_tokens
                    new_targets += [IGNORE_TOKEN_ID] * len(nl_tokens)

                    for j in range(0, best_width, processor["image"].patch_size):
                        new_input_ids += [PATCH_START_ID]
                        new_targets += [IGNORE_TOKEN_ID]

                        image_indice_b = torch.zeros(
                            1, image_token_length, dtype=torch.int64
                        )  # This will change in collate_fn
                        image_indice_s = (
                            torch.arange(
                                len(new_input_ids), len(new_input_ids) + image_token_length
                            )
                            .unsqueeze(0)
                            .repeat(1, 1)
                        )
                        image_indice_b_s = torch.stack(
                            [image_indice_b, image_indice_s], dim=0
                        )  # 2, num_image, image_length
                        image_indices.append(image_indice_b_s)

                        new_input_ids += [PATCH_CONTEXT_ID] * image_token_length
                        new_targets += [IGNORE_TOKEN_ID] * image_token_length

                        new_input_ids += [PATCH_END_ID]
                        new_targets += [IGNORE_TOKEN_ID]

            st = img_pos + 1

        new_input_ids += input_ids[st:]
        new_targets += targets[st:]

        input_ids = new_input_ids
        targets = new_targets

    # ----------------------------------------------------------------
    # video
    if has_video(sample):
        vid_positions = [i for i, x in enumerate(input_ids) if x == VID_TAG_ID]
        assert len(vid_positions) == len(sample["videos"]), sample

        new_input_ids = []
        new_targets = []
        st = 0
        for vid_idx, vid_pos in enumerate(vid_positions):
            video_frames, _ = processor["image"].process_video(
                sample["videos"][vid_idx], max_num_frame, max_fps
            )

            new_input_ids += input_ids[st:vid_pos]
            new_targets += targets[st:vid_pos]

            images.append(video_frames)

            for _ in video_frames:
                new_input_ids += [VID_START_ID]
                new_targets += [IGNORE_TOKEN_ID]

                image_indice_b = torch.zeros(
                    1, image_token_length, dtype=torch.int64
                )  # This will change in collate_fn
                image_indice_s = (
                    torch.arange(len(new_input_ids), len(new_input_ids) + image_token_length)
                    .unsqueeze(0)
                    .repeat(1, 1)
                )
                image_indice_b_s = torch.stack(
                    [image_indice_b, image_indice_s], dim=0
                )  # 2, num_image, image_length
                image_indices.append(image_indice_b_s)

                new_input_ids += [VID_CONTEXT_ID] * image_token_length
                new_targets += [IGNORE_TOKEN_ID] * image_token_length

                new_input_ids += [VID_END_ID]
                new_targets += [IGNORE_TOKEN_ID]

            st = vid_pos + 1

        new_input_ids += input_ids[st:]
        new_targets += targets[st:]

        input_ids = new_input_ids
        targets = new_targets

    # ----------------------------------------------------------------
    # audio
    if has_audio(sample) and processor["audio"].is_contiguous:
        aud_positions = [i for i, x in enumerate(input_ids) if x == AUD_TAG_ID]
        # assert len(aud_positions) == len(sample["audios"]), sample
        assert len(aud_positions) == len(unused_audio_idxs), sample

        new_input_ids = []
        new_targets = []
        st = 0
        for aud_idx, aud_pos in enumerate(aud_positions):
            aud_idx = unused_audio_idxs[aud_idx]
            audio = processor["audio"].process_audios(sample["audios"][aud_idx], is_contiguous=True)
            audios.append(audio)
            audio_token_length = audio.size(0) + 4
            # audio_token_length = audio.size(0)

            new_input_ids += input_ids[st:aud_pos]
            new_targets += targets[st:aud_pos]

            new_input_ids += [AUD_START_ID]
            new_targets += [IGNORE_TOKEN_ID]

            audio_indice_b = torch.zeros(
                1, audio_token_length, dtype=torch.int64
            )  # This will change in collate_fn
            audio_indice_s = (
                torch.arange(len(new_input_ids), len(new_input_ids) + audio_token_length)
                .unsqueeze(0)
                .repeat(1, 1)
            )
            audio_indice_b_s = torch.stack(
                [audio_indice_b, audio_indice_s], dim=0
            )  # 2, num_image, image_length
            audio_indices.append(audio_indice_b_s)

            new_input_ids += [AUD_CONTEXT_ID] * audio_token_length
            new_targets += [IGNORE_TOKEN_ID] * audio_token_length

            new_input_ids += [AUD_END_ID]
            new_targets += [IGNORE_TOKEN_ID]

            st = aud_pos + 1

        new_input_ids += input_ids[st:]
        new_targets += targets[st:]

        input_ids = new_input_ids
        targets = new_targets

    if len(images) > 0:
        images = torch.cat(images, dim=0)

    if len(image_indices) > 0:
        image_indices = torch.cat(image_indices, dim=1)

    # if len(audios) > 0:
    #     audios = torch.cat(audios, dim=0)

    # if len(audio_indices) > 0:
    #     audio_indices = torch.cat(audio_indices, dim=1)

    attention_mask = [1] * len(input_ids)

    # print("sample", sample, flush=True)
    # print("input_ids", input_ids, flush=True)
    # print("targets", targets[:100], flush=True)
    # print("images", [xx.shape for x in images for xx in x], flush=True)

    return dict(
        input_ids=input_ids,
        labels=targets,
        attention_mask=attention_mask,
        images=images,
        image_indices=image_indices,
        audios=audios,
        audio_indices=audio_indices,
    )


def has_video(sample):
    # video
    if (
        "videos" in sample
        and isinstance(sample["videos"], list)
        and None not in sample["videos"]
        and len(sample["videos"])
    ):
        return True
    return False


def has_image(sample):
    # image
    if (
        "images" in sample
        and isinstance(sample["images"], list)
        and None not in sample["images"]
        and len(sample["images"])
    ):
        return True
    return False


def has_audio(sample):
    # audio
    if (
        "audios" in sample
        and isinstance(sample["audios"], list)
        and None not in sample["audios"]
        and len(sample["audios"])
    ):
        return True
    return False


def text_audio_interval_old(input_ids, AUD_START_ID, AUD_END_ID, text_audio_interval_ratio):

    if text_audio_interval_ratio is not None:
        text_num, audio_num = text_audio_interval_ratio

    else:
        text_num = 13
        audio_num = 26
        text_num = 4
        audio_num = 10

    # exclude AUD_START and AUD_END
    audio_num = audio_num - 2

    st = [i for i, x in enumerate(input_ids) if x == AUD_START_ID]
    ed = [i for i, x in enumerate(input_ids) if x == AUD_END_ID]

    # only text
    if len(st) == 0 and len(ed) == 0:
        return input_ids

    assert len(st) == 1
    assert len(ed) == 1

    st = st[0]
    ed = ed[0]

    assert st < ed

    # only audio
    if st == 0 and ed == len(input_ids) - 1:
        return input_ids

    audio_tokens = input_ids[st + 1 : ed]
    text_tokens = input_ids[:st] + input_ids[ed + 1 :]

    if False:
        audio_tokens_chunks = [
            audio_tokens[i : i + audio_num] for i in range(0, len(audio_tokens), audio_num)
        ]
        text_tokens_chunks = [
            text_tokens[i : i + text_num] for i in range(0, len(text_tokens), text_num)
        ]

    if False:
        # [0 1] [2 3 4 5 6 audio_num-1] ...
        audio_tokens_chunks = [audio_tokens[:2], audio_tokens[2:audio_num]] + [
            audio_tokens[i : i + audio_num] for i in range(audio_num, len(audio_tokens), audio_num)
        ]
        # [0] [1 2 text_num-1] ...
        text_tokens_chunks = [text_tokens[:1], text_tokens[1:text_num]] + [
            text_tokens[i : i + text_num] for i in range(text_num, len(text_tokens), text_num)
        ]

    if True:
        # [0 1 2 3 4 5 6 audio_num] [] ...
        audio_tokens_chunks = [audio_tokens[:audio_num]] + [
            audio_tokens[i : i + audio_num] for i in range(audio_num, len(audio_tokens), audio_num)
        ]
        # [0] [] ...
        text_tokens_chunks = [text_tokens[:1]] + [
            text_tokens[i : i + text_num] for i in range(1, len(text_tokens), text_num)
        ]

    chunk_num = min(len(audio_tokens_chunks), len(text_tokens_chunks))
    audio_tokens_chunks = audio_tokens_chunks[: chunk_num - 1] + [
        sum(audio_tokens_chunks[chunk_num - 1 :], [])
    ]
    text_tokens_chunks = text_tokens_chunks[: chunk_num - 1] + [
        sum(text_tokens_chunks[chunk_num - 1 :], [])
    ]

    interval_input_ids = []
    for text_tokens, audio_tokens in zip(text_tokens_chunks, audio_tokens_chunks):
        interval_input_ids += text_tokens + [AUD_START_ID] + audio_tokens + [AUD_END_ID]
        # interval_input_ids += text_tokens + audio_tokens

    return interval_input_ids


def text_audio_interval(input_ids, AUD_START_ID, AUD_END_ID, text_audio_interval_ratio):

    if text_audio_interval_ratio is None:
        #                            T   A
        text_audio_interval_ratio = [13, 26]
        #                            T  A  T  A  T  A
        text_audio_interval_ratio = [1, 4, 3, 8, 4, 10]
        #                            T  A   T  A
        text_audio_interval_ratio = [1, 10, 4, 10]

    text_nums = text_audio_interval_ratio[::2]
    audio_nums = text_audio_interval_ratio[1::2]

    # exclude AUD_START and AUD_END
    audio_nums = [x - 2 for x in audio_nums]

    st = [i for i, x in enumerate(input_ids) if x == AUD_START_ID]
    ed = [i for i, x in enumerate(input_ids) if x == AUD_END_ID]

    # only text
    if len(st) == 0 and len(ed) == 0:
        return input_ids

    assert len(st) == 1
    assert len(ed) == 1

    st = st[0]
    ed = ed[0]

    assert st < ed

    # only audio
    if st == 0 and ed == len(input_ids) - 1:
        return input_ids

    audio_tokens = input_ids[st + 1 : ed]
    text_tokens = input_ids[:st] + input_ids[ed + 1 :]

    audio_tokens_chunks = []
    while len(audio_tokens) > 0:
        if len(audio_nums) > 1:
            audio_num = audio_nums.pop(0)
        else:
            audio_num = audio_nums[0]

        audio_tokens_chunks.append(audio_tokens[:audio_num])
        audio_tokens = audio_tokens[audio_num:]

    text_tokens_chunks = []
    while len(text_tokens) > 0:
        if len(text_nums) > 1:
            text_num = text_nums.pop(0)
        else:
            text_num = text_nums[0]

        text_tokens_chunks.append(text_tokens[:text_num])
        text_tokens = text_tokens[text_num:]

    chunk_num = min(len(audio_tokens_chunks), len(text_tokens_chunks))
    audio_tokens_chunks = audio_tokens_chunks[: chunk_num - 1] + [
        sum(audio_tokens_chunks[chunk_num - 1 :], [])
    ]
    text_tokens_chunks = text_tokens_chunks[: chunk_num - 1] + [
        sum(text_tokens_chunks[chunk_num - 1 :], [])
    ]

    interval_input_ids = []
    for text_tokens, audio_tokens in zip(text_tokens_chunks, audio_tokens_chunks):
        interval_input_ids += text_tokens + [AUD_START_ID] + audio_tokens + [AUD_END_ID]
        # interval_input_ids += text_tokens + audio_tokens

    return interval_input_ids



from .base import ImageGenProcessor


def get_image_gen_processor(tag, **kwargs) -> ImageGenProcessor:
    if tag == "HFApiInferenceImageGenProcessor":
        from .hf_img_gen_processor import HFApiInferenceImageGenProcessor

        return HFApiInferenceImageGenProcessor(**kwargs)
    if tag == "ComfyUIAPIImageGenProcessor":
        from .comfyui_img_gen_processor import ComfyUIAPIImageGenProcessor

        return ComfyUIAPIImageGenProcessor(**kwargs)
    if tag == "OpenAIImageGenProcessor":
        from .openai_img_gen_processor import OpenAIImageGenProcessor

        return OpenAIImageGenProcessor(**kwargs)
    if tag == "TogetherImageGenProcessor":
        from .together_img_gen_processor import TogetherImageGenProcessor

        return TogetherImageGenProcessor(**kwargs)
    if tag == "HFStableDiffusionImageGenProcessor":
        from .diffusers_img_gen_processor import HFStableDiffusionImageGenProcessor

        return HFStableDiffusionImageGenProcessor(**kwargs)



"""Example script for generating audio using HiggsAudio."""

import click
import soundfile as sf
import langid
import jieba
import os
import re
import copy
import torchaudio
import tqdm
import yaml

from loguru import logger
from boson_multimodal.serve.serve_engine import HiggsAudioServeEngine, HiggsAudioResponse
from boson_multimodal.data_types import Message, ChatMLSample, AudioContent, TextContent

from boson_multimodal.model.higgs_audio import HiggsAudioConfig, HiggsAudioModel
from boson_multimodal.data_collator.higgs_audio_collator import HiggsAudioSampleCollator
from boson_multimodal.audio_processing.higgs_audio_tokenizer import load_higgs_audio_tokenizer
from boson_multimodal.dataset.chatml_dataset import (
    ChatMLDatasetSample,
    prepare_chatml_sample,
)
from boson_multimodal.model.higgs_audio.utils import revert_delay_pattern
from typing import List
from transformers import AutoConfig, AutoTokenizer
from transformers.cache_utils import StaticCache
from typing import Optional
from dataclasses import asdict
import torch

CURR_DIR = os.path.dirname(os.path.abspath(__file__))


AUDIO_PLACEHOLDER_TOKEN = "<|__AUDIO_PLACEHOLDER__|>"


MULTISPEAKER_DEFAULT_SYSTEM_MESSAGE = """You are an AI assistant designed to convert text into speech.
If the user's message includes a [SPEAKER*] tag, do not read out the tag and generate speech for the following text, using the specified voice.
If no speaker tag is present, select a suitable voice on your own."""


def normalize_chinese_punctuation(text):
    """
    Convert Chinese (full-width) punctuation marks to English (half-width) equivalents.
    """
    # Mapping of Chinese punctuation to English punctuation
    chinese_to_english_punct = {
        "，": ", ",  # comma
        "。": ".",  # period
        "：": ":",  # colon
        "；": ";",  # semicolon
        "？": "?",  # question mark
        "！": "!",  # exclamation mark
        "（": "(",  # left parenthesis
        "）": ")",  # right parenthesis
        "【": "[",  # left square bracket
        "】": "]",  # right square bracket
        "《": "<",  # left angle quote
        "》": ">",  # right angle quote
        "“": '"',  # left double quotation
        "”": '"',  # right double quotation
        "‘": "'",  # left single quotation
        "’": "'",  # right single quotation
        "、": ",",  # enumeration comma
        "—": "-",  # em dash
        "…": "...",  # ellipsis
        "·": ".",  # middle dot
        "「": '"',  # left corner bracket
        "」": '"',  # right corner bracket
        "『": '"',  # left double corner bracket
        "』": '"',  # right double corner bracket
    }

    # Replace each Chinese punctuation with its English counterpart
    for zh_punct, en_punct in chinese_to_english_punct.items():
        text = text.replace(zh_punct, en_punct)

    return text


def prepare_chunk_text(
    text, chunk_method: Optional[str] = None, chunk_max_word_num: int = 100, chunk_max_num_turns: int = 1
):
    """Chunk the text into smaller pieces. We will later feed the chunks one by one to the model.

    Parameters
    ----------
    text : str
        The text to be chunked.
    chunk_method : str, optional
        The method to use for chunking. Options are "speaker", "word", or None. By default, we won't use any chunking and
        will feed the whole text to the model.
    replace_speaker_tag_with_special_tags : bool, optional
        Whether to replace speaker tags with special tokens, by default False
        If the flag is set to True, we will replace [SPEAKER0] with <|speaker_id_start|>SPEAKER0<|speaker_id_end|>
    chunk_max_word_num : int, optional
        The maximum number of words for each chunk when "word" chunking method is used, by default 100
    chunk_max_num_turns : int, optional
        The maximum number of turns for each chunk when "speaker" chunking method is used,

    Returns
    -------
    List[str]
        The list of text chunks.

    """
    if chunk_method is None:
        return [text]
    elif chunk_method == "speaker":
        lines = text.split("\n")
        speaker_chunks = []
        speaker_utterance = ""
        for line in lines:
            line = line.strip()
            if line.startswith("[SPEAKER") or line.startswith("<|speaker_id_start|>"):
                if speaker_utterance:
                    speaker_chunks.append(speaker_utterance.strip())
                speaker_utterance = line
            else:
                if speaker_utterance:
                    speaker_utterance += "\n" + line
                else:
                    speaker_utterance = line
        if speaker_utterance:
            speaker_chunks.append(speaker_utterance.strip())
        if chunk_max_num_turns > 1:
            merged_chunks = []
            for i in range(0, len(speaker_chunks), chunk_max_num_turns):
                merged_chunk = "\n".join(speaker_chunks[i : i + chunk_max_num_turns])
                merged_chunks.append(merged_chunk)
            return merged_chunks
        return speaker_chunks
    elif chunk_method == "word":
        # TODO: We may improve the logic in the future
        # For long-form generation, we will first divide the text into multiple paragraphs by splitting with "\n\n"
        # After that, we will chunk each paragraph based on word count
        language = langid.classify(text)[0]
        paragraphs = text.split("\n\n")
        chunks = []
        for idx, paragraph in enumerate(paragraphs):
            if language == "zh":
                # For Chinese, we will chunk based on character count
                words = list(jieba.cut(paragraph, cut_all=False))
                for i in range(0, len(words), chunk_max_word_num):
                    chunk = "".join(words[i : i + chunk_max_word_num])
                    chunks.append(chunk)
            else:
                words = paragraph.split(" ")
                for i in range(0, len(words), chunk_max_word_num):
                    chunk = " ".join(words[i : i + chunk_max_word_num])
                    chunks.append(chunk)
            chunks[-1] += "\n\n"
        return chunks
    else:
        raise ValueError(f"Unknown chunk method: {chunk_method}")


def _build_system_message_with_audio_prompt(system_message):
    contents = []

    while AUDIO_PLACEHOLDER_TOKEN in system_message:
        loc = system_message.find(AUDIO_PLACEHOLDER_TOKEN)
        contents.append(TextContent(system_message[:loc]))
        contents.append(AudioContent(audio_url=""))
        system_message = system_message[loc + len(AUDIO_PLACEHOLDER_TOKEN) :]

    if len(system_message) > 0:
        contents.append(TextContent(system_message))
    ret = Message(
        role="system",
        content=contents,
    )
    return ret


class HiggsAudioModelClient:
    def __init__(
        self,
        model_path,
        audio_tokenizer,
        device_id=None,
        max_new_tokens=2048,
        kv_cache_lengths: List[int] = [1024, 4096, 8192],  # Multiple KV cache sizes,
        use_static_kv_cache=False,
    ):
        if device_id is None:
            self._device = "cuda" if torch.cuda.is_available() else "cpu"
        else:
            self._device = f"cuda:{device_id}"
        self._audio_tokenizer = (
            load_higgs_audio_tokenizer(audio_tokenizer, device=self._device)
            if isinstance(audio_tokenizer, str)
            else audio_tokenizer
        )
        self._model = HiggsAudioModel.from_pretrained(
            model_path,
            device_map=self._device,
            torch_dtype=torch.bfloat16,
        )
        self._model.eval()
        self._kv_cache_lengths = kv_cache_lengths
        self._use_static_kv_cache = use_static_kv_cache

        self._tokenizer = AutoTokenizer.from_pretrained(model_path)
        self._config = AutoConfig.from_pretrained(model_path)
        self._max_new_tokens = max_new_tokens
        self._collator = HiggsAudioSampleCollator(
            whisper_processor=None,
            audio_in_token_id=self._config.audio_in_token_idx,
            audio_out_token_id=self._config.audio_out_token_idx,
            audio_stream_bos_id=self._config.audio_stream_bos_id,
            audio_stream_eos_id=self._config.audio_stream_eos_id,
            encode_whisper_embed=self._config.encode_whisper_embed,
            pad_token_id=self._config.pad_token_id,
            return_audio_in_tokens=self._config.encode_audio_in_tokens,
            use_delay_pattern=self._config.use_delay_pattern,
            round_to=1,
            audio_num_codebooks=self._config.audio_num_codebooks,
        )
        self.kv_caches = None
        if use_static_kv_cache:
            self._init_static_kv_cache()

    def _init_static_kv_cache(self):
        cache_config = copy.deepcopy(self._model.config.text_config)
        cache_config.num_hidden_layers = self._model.config.text_config.num_hidden_layers
        if self._model.config.audio_dual_ffn_layers:
            cache_config.num_hidden_layers += len(self._model.config.audio_dual_ffn_layers)
        # A list of KV caches for different lengths
        self.kv_caches = {
            length: StaticCache(
                config=cache_config,
                max_batch_size=1,
                max_cache_len=length,
                device=self._model.device,
                dtype=self._model.dtype,
            )
            for length in sorted(self._kv_cache_lengths)
        }
        # Capture CUDA graphs for each KV cache length
        if "cuda" in self._device:
            logger.info(f"Capturing CUDA graphs for each KV cache length")
            self._model.capture_model(self.kv_caches.values())

    def _prepare_kv_caches(self):
        for kv_cache in self.kv_caches.values():
            kv_cache.reset()

    @torch.inference_mode()
    def generate(
        self,
        messages,
        audio_ids,
        chunked_text,
        generation_chunk_buffer_size,
        temperature=1.0,
        top_k=50,
        top_p=0.95,
        ras_win_len=7,
        ras_win_max_num_repeat=2,
        seed=123,
        *args,
        **kwargs,
    ):
        if ras_win_len is not None and ras_win_len <= 0:
            ras_win_len = None
        sr = 24000
        audio_out_ids_l = []
        generated_audio_ids = []
        generation_messages = []
        for idx, chunk_text in tqdm.tqdm(
            enumerate(chunked_text), desc="Generating audio chunks", total=len(chunked_text)
        ):
            generation_messages.append(
                Message(
                    role="user",
                    content=chunk_text,
                )
            )
            chatml_sample = ChatMLSample(messages=messages + generation_messages)
            input_tokens, _, _, _ = prepare_chatml_sample(chatml_sample, self._tokenizer)
            postfix = self._tokenizer.encode(
                "<|start_header_id|>assistant<|end_header_id|>\n\n", add_special_tokens=False
            )
            input_tokens.extend(postfix)

            logger.info(f"========= Chunk {idx} Input =========")
            logger.info(self._tokenizer.decode(input_tokens))
            context_audio_ids = audio_ids + generated_audio_ids

            curr_sample = ChatMLDatasetSample(
                input_ids=torch.LongTensor(input_tokens),
                label_ids=None,
                audio_ids_concat=torch.concat([ele.cpu() for ele in context_audio_ids], dim=1)
                if context_audio_ids
                else None,
                audio_ids_start=torch.cumsum(
                    torch.tensor([0] + [ele.shape[1] for ele in context_audio_ids], dtype=torch.long), dim=0
                )
                if context_audio_ids
                else None,
                audio_waveforms_concat=None,
                audio_waveforms_start=None,
                audio_sample_rate=None,
                audio_speaker_indices=None,
            )

            batch_data = self._collator([curr_sample])
            batch = asdict(batch_data)
            for k, v in batch.items():
                if isinstance(v, torch.Tensor):
                    batch[k] = v.contiguous().to(self._device)

            if self._use_static_kv_cache:
                self._prepare_kv_caches()

            # Generate audio
            outputs = self._model.generate(
                **batch,
                max_new_tokens=self._max_new_tokens,
                use_cache=True,
                do_sample=True,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                past_key_values_buckets=self.kv_caches,
                ras_win_len=ras_win_len,
                ras_win_max_num_repeat=ras_win_max_num_repeat,
                stop_strings=["<|end_of_text|>", "<|eot_id|>"],
                tokenizer=self._tokenizer,
                seed=seed,
            )

            step_audio_out_ids_l = []
            for ele in outputs[1]:
                audio_out_ids = ele
                if self._config.use_delay_pattern:
                    audio_out_ids = revert_delay_pattern(audio_out_ids)
                step_audio_out_ids_l.append(audio_out_ids.clip(0, self._audio_tokenizer.codebook_size - 1)[:, 1:-1])
            audio_out_ids = torch.concat(step_audio_out_ids_l, dim=1)
            audio_out_ids_l.append(audio_out_ids)
            generated_audio_ids.append(audio_out_ids)

            generation_messages.append(
                Message(
                    role="assistant",
                    content=AudioContent(audio_url=""),
                )
            )
            if generation_chunk_buffer_size is not None and len(generated_audio_ids) > generation_chunk_buffer_size:
                generated_audio_ids = generated_audio_ids[-generation_chunk_buffer_size:]
                generation_messages = generation_messages[(-2 * generation_chunk_buffer_size) :]

        logger.info(f"========= Final Text output =========")
        logger.info(self._tokenizer.decode(outputs[0][0]))
        concat_audio_out_ids = torch.concat(audio_out_ids_l, dim=1)
        concat_wv = self._audio_tokenizer.decode(concat_audio_out_ids.unsqueeze(0))[0, 0]
        text_result = self._tokenizer.decode(outputs[0][0])
        return concat_wv, sr, text_result


def prepare_generation_context(scene_prompt, ref_audio, ref_audio_in_system_message, audio_tokenizer, speaker_tags):
    """Prepare the context for generation.

    The context contains the system message, user message, assistant message, and audio prompt if any.
    """
    system_message = None
    messages = []
    audio_ids = []
    if ref_audio is not None:
        num_speakers = len(ref_audio.split(","))
        speaker_info_l = ref_audio.split(",")
        voice_profile = None
        if any([speaker_info.startswith("profile:") for speaker_info in ref_audio.split(",")]):
            ref_audio_in_system_message = True
        if ref_audio_in_system_message:
            speaker_desc = []
            for spk_id, character_name in enumerate(speaker_info_l):
                if character_name.startswith("profile:"):
                    if voice_profile is None:
                        with open(f"{CURR_DIR}/voice_prompts/profile.yaml", "r", encoding="utf-8") as f:
                            voice_profile = yaml.safe_load(f)
                    character_desc = voice_profile["profiles"][character_name[len("profile:") :].strip()]
                    speaker_desc.append(f"SPEAKER{spk_id}: {character_desc}")
                else:
                    speaker_desc.append(f"SPEAKER{spk_id}: {AUDIO_PLACEHOLDER_TOKEN}")
            if scene_prompt:
                system_message = (
                    "Generate audio following instruction."
                    "\n\n"
                    f"<|scene_desc_start|>\n{scene_prompt}\n\n" + "\n".join(speaker_desc) + "\n<|scene_desc_end|>"
                )
            else:
                system_message = (
                    "Generate audio following instruction.\n\n"
                    + f"<|scene_desc_start|>\n"
                    + "\n".join(speaker_desc)
                    + "\n<|scene_desc_end|>"
                )
            system_message = _build_system_message_with_audio_prompt(system_message)
        else:
            if scene_prompt:
                system_message = Message(
                    role="system",
                    content=f"Generate audio following instruction.\n\n<|scene_desc_start|>\n{scene_prompt}\n<|scene_desc_end|>",
                )
        voice_profile = None
        for spk_id, character_name in enumerate(ref_audio.split(",")):
            if not character_name.startswith("profile:"):
                prompt_audio_path = os.path.join(f"{CURR_DIR}/voice_prompts", f"{character_name}.wav")
                prompt_text_path = os.path.join(f"{CURR_DIR}/voice_prompts", f"{character_name}.txt")
                assert os.path.exists(prompt_audio_path), (
                    f"Voice prompt audio file {prompt_audio_path} does not exist."
                )
                assert os.path.exists(prompt_text_path), f"Voice prompt text file {prompt_text_path} does not exist."
                with open(prompt_text_path, "r", encoding="utf-8") as f:
                    prompt_text = f.read().strip()
                audio_tokens = audio_tokenizer.encode(prompt_audio_path)
                audio_ids.append(audio_tokens)

                if not ref_audio_in_system_message:
                    messages.append(
                        Message(
                            role="user",
                            content=f"[SPEAKER{spk_id}] {prompt_text}" if num_speakers > 1 else prompt_text,
                        )
                    )
                    messages.append(
                        Message(
                            role="assistant",
                            content=AudioContent(
                                audio_url=prompt_audio_path,
                            ),
                        )
                    )
    else:
        if len(speaker_tags) > 1:
            # By default, we just alternate between male and female voices
            speaker_desc_l = []

            for idx, tag in enumerate(speaker_tags):
                if idx % 2 == 0:
                    speaker_desc = f"feminine"
                else:
                    speaker_desc = f"masculine"
                speaker_desc_l.append(f"{tag}: {speaker_desc}")

            speaker_desc = "\n".join(speaker_desc_l)
            scene_desc_l = []
            if scene_prompt:
                scene_desc_l.append(scene_prompt)
            scene_desc_l.append(speaker_desc)
            scene_desc = "\n\n".join(scene_desc_l)

            system_message = Message(
                role="system",
                content=f"{MULTISPEAKER_DEFAULT_SYSTEM_MESSAGE}\n\n<|scene_desc_start|>\n{scene_desc}\n<|scene_desc_end|>",
            )
        else:
            system_message_l = ["Generate audio following instruction."]
            if scene_prompt:
                system_message_l.append(f"<|scene_desc_start|>\n{scene_prompt}\n<|scene_desc_end|>")
            system_message = Message(
                role="system",
                content="\n\n".join(system_message_l),
            )
    if system_message:
        messages.insert(0, system_message)
    return messages, audio_ids


@click.command()
@click.option(
    "--model_path",
    type=str,
    default="bosonai/higgs-audio-v2-generation-3B-base",
    help="Output wav file path.",
)
@click.option(
    "--audio_tokenizer",
    type=str,
    default="bosonai/higgs-audio-v2-tokenizer",
    help="Audio tokenizer path, if not set, use the default one.",
)
@click.option(
    "--max_new_tokens",
    type=int,
    default=2048,
    help="The maximum number of new tokens to generate.",
)
@click.option(
    "--transcript",
    type=str,
    default="transcript/single_speaker/en_dl.txt",
    help="The prompt to use for generation. If not set, we will use a default prompt.",
)
@click.option(
    "--scene_prompt",
    type=str,
    default=f"{CURR_DIR}/scene_prompts/quiet_indoor.txt",
    help="The scene description prompt to use for generation. If not set, or set to `empty`, we will leave it to empty.",
)
@click.option(
    "--temperature",
    type=float,
    default=1.0,
    help="The value used to module the next token probabilities.",
)
@click.option(
    "--top_k",
    type=int,
    default=50,
    help="The number of highest probability vocabulary tokens to keep for top-k-filtering.",
)
@click.option(
    "--top_p",
    type=float,
    default=0.95,
    help="If set to float < 1, only the most probable tokens with probabilities that add up to top_p or higher are kept for generation.",
)
@click.option(
    "--ras_win_len",
    type=int,
    default=7,
    help="The window length for RAS sampling. If set to 0 or a negative value, we won't use RAS sampling.",
)
@click.option(
    "--ras_win_max_num_repeat",
    type=int,
    default=2,
    help="The maximum number of times to repeat the RAS window. Only used when --ras_win_len is set.",
)
@click.option(
    "--ref_audio",
    type=str,
    default=None,
    help="The voice prompt to use for generation. If not set, we will let the model randomly pick a voice. "
    "For multi-speaker generation, you can specify the prompts as `belinda,chadwick` and we will use the voice of belinda as SPEAKER0 and the voice of chadwick as SPEAKER1.",
)
@click.option(
    "--ref_audio_in_system_message",
    is_flag=True,
    default=False,
    help="Whether to include the voice prompt description in the system message.",
    show_default=True,
)
@click.option(
    "--chunk_method",
    default=None,
    type=click.Choice([None, "speaker", "word"]),
    help="The method to use for chunking the prompt text. Options are 'speaker', 'word', or None. By default, we won't use any chunking and will feed the whole text to the model.",
)
@click.option(
    "--chunk_max_word_num",
    default=200,
    type=int,
    help="The maximum number of words for each chunk when 'word' chunking method is used. Only used when --chunk_method is set to 'word'.",
)
@click.option(
    "--chunk_max_num_turns",
    default=1,
    type=int,
    help="The maximum number of turns for each chunk when 'speaker' chunking method is used. Only used when --chunk_method is set to 'speaker'.",
)
@click.option(
    "--generation_chunk_buffer_size",
    default=None,
    type=int,
    help="The maximal number of chunks to keep in the buffer. We will always keep the reference audios, and keep `max_chunk_buffer` chunks of generated audio.",
)
@click.option(
    "--seed",
    default=None,
    type=int,
    help="Random seed for generation.",
)
@click.option(
    "--device_id",
    type=int,
    default=None,
    help="The device to run the model on.",
)
@click.option(
    "--out_path",
    type=str,
    default="generation.wav",
)
@click.option(
    "--use_static_kv_cache",
    type=int,
    default=1,
    help="Whether to use static KV cache for faster generation. Only works when using GPU.",
)
def main(
    model_path,
    audio_tokenizer,
    max_new_tokens,
    transcript,
    scene_prompt,
    temperature,
    top_k,
    top_p,
    ras_win_len,
    ras_win_max_num_repeat,
    ref_audio,
    ref_audio_in_system_message,
    chunk_method,
    chunk_max_word_num,
    chunk_max_num_turns,
    generation_chunk_buffer_size,
    seed,
    device_id,
    out_path,
    use_static_kv_cache,
):
    gen(
        model_path,
        audio_tokenizer,
        max_new_tokens,
        transcript,
        scene_prompt,
        temperature,
        top_k,
        top_p,
        ras_win_len,
        ras_win_max_num_repeat,
        ref_audio,
        ref_audio_in_system_message,
        chunk_method,
        chunk_max_word_num,
        chunk_max_num_turns,
        generation_chunk_buffer_size,
        seed,
        device_id,
        out_path,
        use_static_kv_cache,
    )


def gen(
    model_path,
    audio_tokenizer,
    max_new_tokens,
    transcript,
    scene_prompt,
    temperature,
    top_k,
    top_p,
    ras_win_len,
    ras_win_max_num_repeat,
    ref_audio,
    ref_audio_in_system_message,
    chunk_method,
    chunk_max_word_num,
    chunk_max_num_turns,
    generation_chunk_buffer_size,
    seed,
    device_id,
    out_path,
    use_static_kv_cache,
):
    if device_id is None:
        if torch.cuda.is_available():
            device_id = 0
            device = "cuda:0"
        else:
            device_id = None
            device = "cpu"
    else:
        device = f"cuda:{device_id}"
    audio_tokenizer = load_higgs_audio_tokenizer(audio_tokenizer, device=device)

    model_client = HiggsAudioModelClient(
        model_path=model_path,
        audio_tokenizer=audio_tokenizer,
        device_id=device_id,
        max_new_tokens=max_new_tokens,
        use_static_kv_cache=use_static_kv_cache,
    )
    pattern = re.compile(r"\[(SPEAKER\d+)\]")

    if os.path.exists(transcript):
        logger.info(f"Loading transcript from {transcript}")
        with open(transcript, "r", encoding="utf-8") as f:
            transcript = f.read().strip()

    if scene_prompt is not None and scene_prompt != "empty" and os.path.exists(scene_prompt):
        with open(scene_prompt, "r", encoding="utf-8") as f:
            scene_prompt = f.read().strip()
    else:
        scene_prompt = None

    speaker_tags = sorted(set(pattern.findall(transcript)))

    # Perform some basic normalization
    transcript = normalize_chinese_punctuation(transcript)
    # Other normalizations (e.g., parentheses and other symbols. Will be improved in the future)
    transcript = transcript.replace("(", " ")
    transcript = transcript.replace(")", " ")
    transcript = transcript.replace("°F", " degrees Fahrenheit")
    transcript = transcript.replace("°C", " degrees Celsius")

    for tag, replacement in [
        ("[laugh]", "<SE>[Laughter]</SE>"),
        ("[humming start]", "<SE>[Humming]</SE>"),
        ("[humming end]", "<SE_e>[Humming]</SE_e>"),
        ("[music start]", "<SE_s>[Music]</SE_s>"),
        ("[music end]", "<SE_e>[Music]</SE_e>"),
        ("[music]", "<SE>[Music]</SE>"),
        ("[sing start]", "<SE_s>[Singing]</SE_s>"),
        ("[sing end]", "<SE_e>[Singing]</SE_e>"),
        ("[applause]", "<SE>[Applause]</SE>"),
        ("[cheering]", "<SE>[Cheering]</SE>"),
        ("[cough]", "<SE>[Cough]</SE>"),
    ]:
        transcript = transcript.replace(tag, replacement)
    lines = transcript.split("\n")
    transcript = "\n".join([" ".join(line.split()) for line in lines if line.strip()])
    transcript = transcript.strip()

    if not any([transcript.endswith(c) for c in [".", "!", "?", ",", ";", '"', "'", "</SE_e>", "</SE>"]]):
        transcript += "."

    messages, audio_ids = prepare_generation_context(
        scene_prompt=scene_prompt,
        ref_audio=ref_audio,
        ref_audio_in_system_message=ref_audio_in_system_message,
        audio_tokenizer=audio_tokenizer,
        speaker_tags=speaker_tags,
    )
    chunked_text = prepare_chunk_text(
        transcript,
        chunk_method=chunk_method,
        chunk_max_word_num=chunk_max_word_num,
        chunk_max_num_turns=chunk_max_num_turns,
    )

    logger.info("Chunks used for generation:")
    for idx, chunk_text in enumerate(chunked_text):
        logger.info(f"Chunk {idx}:")
        logger.info(chunk_text)
        logger.info("-----")

    concat_wv, sr, text_output = model_client.generate(
        messages=messages,
        audio_ids=audio_ids,
        chunked_text=chunked_text,
        generation_chunk_buffer_size=generation_chunk_buffer_size,
        temperature=temperature,
        top_k=top_k,
        top_p=top_p,
        ras_win_len=ras_win_len,
        ras_win_max_num_repeat=ras_win_max_num_repeat,
        seed=seed,
    )

    sf.write(out_path, concat_wv, sr)
    logger.info(f"{text_output=}")
    logger.info(f"Wav file is saved to '{out_path}' with sample rate {sr}")


if __name__ == "__main__":
    main()



weight = 'pretrained_models/lam_audio2exp.tar'  # path to model weight
ex_vol = True # Isolates vocal track from audio file
audio_input = './assets/sample_audio/BarackObama.wav'
save_json_path = 'bsData.json'

audio_sr = 16000
fps = 30.0

movement_smooth = True
brow_movement = True
id_idx = 153

resume = False  # whether to resume training process
evaluate = True  # evaluate after each epoch training process
test_only = False  # test process

seed = None  # train process will init a random seed and record
save_path = "exp/audio2exp"
num_worker = 16  # total worker in all gpu
batch_size = 16  # total batch size in all gpu
batch_size_val = None  # auto adapt to bs 1 for each gpu
batch_size_test = None  # auto adapt to bs 1 for each gpu
epoch = 100  # total epoch, data loop = epoch // eval_epoch
eval_epoch = 100  # sche total eval & checkpoint epoch

sync_bn = False
enable_amp = False
empty_cache = False
find_unused_parameters = False

mix_prob = 0
param_dicts = None  # example: param_dicts = [dict(keyword="block", lr_scale=0.1)]

# model settings
model = dict(
    type="DefaultEstimator",
    backbone=dict(
        type="Audio2Expression",
        pretrained_encoder_type='wav2vec',
        pretrained_encoder_path='facebook/wav2vec2-base-960h',
        wav2vec2_config_path = 'configs/wav2vec2_config.json',
        num_identity_classes=5016,
        identity_feat_dim=64,
        hidden_dim=512,
        expression_dim=52,
        norm_type='ln',
        use_transformer=True,
        num_attention_heads=8,
        num_transformer_layers=6,
    ),
    criteria=[dict(type="L1Loss", loss_weight=1.0, ignore_index=-1)],
)

dataset_type = 'audio2exp'
data_root = './'
data = dict(
    train=dict(
        type=dataset_type,
        split="train",
        data_root=data_root,
        test_mode=False,
    ),
    val=dict(
        type=dataset_type,
        split="val",
        data_root=data_root,
        test_mode=False,
    ),
    test=dict(
        type=dataset_type,
        split="val",
        data_root=data_root,
        test_mode=True
        ),
)

# hook
hooks = [
    dict(type="CheckpointLoader"),
    dict(type="IterationTimer", warmup_iter=2),
    dict(type="InformationWriter"),
    dict(type="SemSegEvaluator"),
    dict(type="CheckpointSaver", save_freq=None),
    dict(type="PreciseEvaluator", test_last=False),
]

# Trainer
train = dict(type="DefaultTrainer")

# Tester
infer = dict(type="Audio2ExpressionInfer",
             verbose=True)



import time

from clearvoice import ClearVoice
import typer

app = typer.Typer()


@app.command()
def offline():
    myClearVoice = ClearVoice(task="speech_enhancement", model_names=["FRCRN_SE_16K"])

    # process single wave file
    start = time.time()
    output_wav = myClearVoice(input_path="records/speech_with_noise16k.wav", online_write=False)
    print(f"cost: {time.time() - start} s")
    myClearVoice.write(output_wav, output_path="records/output_FRCRN_SE_16K.wav")


"""
python demo/clearvoice/FRCRN_SE_16K.py
"""
if __name__ == "__main__":
    import torch
    import os

    thread_count = os.cpu_count()
    print(f"thread_count: {thread_count}")

    torch.set_num_threads(min(8, thread_count))
    torch.set_num_interop_threads(min(8, thread_count))

    app()



import os
import modal


app = modal.App("train-demo-pp")

demo_image = modal.Image.debian_slim(python_version="3.12").pip_install("torch")

MOCK_EPOCH = 5  # 模拟 5 个 epoch


@app.function(
    image=demo_image,
    gpu=os.getenv("IMAGE_GPU", "T4:2"),
    retries=0,
    # how long should we stay up with no requests?
    scaledown_window=15 * 60,
)
def run(mode):
    import torch
    import torch.multiprocessing as mp

    world_size = torch.cuda.device_count()  # 使用所有可用 GPU
    if world_size < 2:
        print("需要至少 2 个 GPU 来演示张量并行")
        return

    if mode == "async":
        train_async_pipeline_parallel(world_size)
    else:
        train_pipeline_parallel(world_size)


"""
modal run src/train/demo/pp.py
modal run src/train/demo/pp.py --mode async
"""


@app.local_entrypoint()
def main(mode=""):
    run.remote(mode)


# 训练函数
def train_pipeline_parallel(world_size):
    import os
    import torch
    import torch.nn as nn
    import torch.optim as optim
    import torch.distributed as dist
    from torch.multiprocessing import Process

    # 初始化分布式环境
    def setup_distributed(rank, world_size):
        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "12355"
        dist.init_process_group("nccl", rank=rank, world_size=world_size)
        torch.cuda.set_device(rank)

    # 清理分布式环境
    def cleanup():
        dist.destroy_process_group()

    # 定义模型的阶段
    class Stage1(nn.Module):
        def __init__(self, input_size, hidden_size):
            super(Stage1, self).__init__()
            self.fc1 = nn.Linear(input_size, hidden_size)
            self.relu = nn.ReLU()

        def forward(self, x):  # [batch_size, input_size]
            return self.relu(self.fc1(x))  # [batch_size, hidden_size]

    class Stage2(nn.Module):
        def __init__(self, hidden_size, output_size):
            super(Stage2, self).__init__()
            self.fc2 = nn.Linear(hidden_size, output_size)

        def forward(self, x):  # [batch_size, hidden_size]
            return self.fc2(x)  # [batch_size, output_size]

    # 流水线训练函数
    def train_pipeline(rank, world_size, data, labels):
        # 设置分布式环境
        setup_distributed(rank, world_size)

        # 参数设置
        input_size = 16
        hidden_size = 32
        output_size = 2
        device = torch.device(f"cuda:{rank}")

        # 将数据和标签移动到对应设备
        data, labels = data.to(device), labels.to(device)

        # 根据 rank 分配阶段
        if rank == 0:
            model = Stage1(input_size, hidden_size).to(device)
        elif rank == 1:
            model = Stage2(hidden_size, output_size).to(device)

        # 定义优化器
        optimizer = torch.optim.SGD(model.parameters(), lr=0.01)

        # 训练步骤
        num_microbatches = 4  # 微批次数量
        microbatch_size = data.size(0) // num_microbatches

        for epoch in range(MOCK_EPOCH):
            optimizer.zero_grad()
            microbatch_size = data.size(0) // num_microbatches

            # 前向传播：微批次处理
            intermediates = []
            outputs = []
            for i in range(num_microbatches):
                start = i * microbatch_size
                end = start + microbatch_size
                micro_data = data[start:end]
                micro_labels = labels[start:end]

                if rank == 0:
                    # 第一阶段：计算并发送中间结果
                    intermediate = model(micro_data)
                    intermediate = intermediate.detach().requires_grad_(True)  # 保留梯度
                    dist.send(tensor=intermediate, dst=1)
                    intermediates.append(intermediate)
                elif rank == 1:
                    # 第二阶段：接收中间结果并计算输出
                    intermediate = torch.zeros(
                        microbatch_size, hidden_size, device=device, requires_grad=True
                    )  # 注意这里需要申请要求保留梯度, 因为需要反向传播计算梯度
                    dist.recv(tensor=intermediate, src=0)
                    output = model(intermediate)  # 确保 intermediate 参与计算图
                    outputs.append((output, micro_labels, intermediate))

            # 反向传播：计算损失并传递梯度
            # 最后一步用来计算loss, 反向传播后，把计算的梯度发送给上一阶段来反向传播计算梯度，依次类推
            if rank == 1:
                loss_fn = nn.CrossEntropyLoss(reduction="sum")  # 使用 sum 以累积微批次损失
                total_loss = 0
                for output, micro_labels, intermediate in outputs:
                    # print(f"output: {output} {output.shape}, {micro_labels} {micro_labels.shape}")
                    loss = loss_fn(output, micro_labels)
                    total_loss += loss
                    loss.backward()  # 为每个微批次单独反向传播，计算梯度,确保 intermediate.grad 生成

                    # 发送中间梯度给 Stage 1
                    assert intermediate.grad is not None, "intermediate.grad is None in Stage 2"
                    dist.send(tensor=intermediate.grad, dst=0)

                total_loss /= num_microbatches  # 平均损失
                print(f"Epoch {epoch+1} loss: {total_loss}")
            elif rank == 0:
                for intermediate in intermediates:
                    grad = torch.zeros_like(intermediate).to(device)
                    dist.recv(tensor=grad, src=1)
                    intermediate.backward(grad)  # 应用梯度

            optimizer.step()

            # 同步并打印损失
            dist.barrier()
            if rank == 0:
                print(f"Epoch {epoch+1} finished on Rank {rank}")
        cleanup()

    # 生成模拟数据
    batch_size = 16
    input_size = 16
    output_size = 2  # 二分类
    data = torch.randn(batch_size, input_size, requires_grad=True)
    labels = torch.randint(0, output_size, (batch_size,))

    print(f"Running Pipeline Parallelism with {world_size} GPUs")

    # 启动进程
    processes = []
    for rank in range(world_size):
        p = Process(target=train_pipeline, args=(rank, world_size, data, labels))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()


def train_async_pipeline_parallel(world_size):
    import torch
    import torch.nn as nn
    import torch.distributed as dist
    from torch.multiprocessing import Process, Queue
    import os
    import queue
    from datetime import timedelta

    # 初始化分布式环境
    def setup_distributed(rank, world_size):
        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "12355"
        dist.init_process_group("nccl", rank=rank, world_size=world_size)
        torch.cuda.set_device(rank)

    # 清理分布式环境
    def cleanup():
        dist.destroy_process_group()

    # 定义模型的阶段
    class Stage1(nn.Module):
        def __init__(self, input_size, hidden_size):
            super(Stage1, self).__init__()
            self.fc1 = nn.Linear(input_size, hidden_size)
            self.relu = nn.ReLU()

        def forward(self, x):
            return self.relu(self.fc1(x))

    class Stage2(nn.Module):
        def __init__(self, hidden_size, output_size):
            super(Stage2, self).__init__()
            self.fc2 = nn.Linear(hidden_size, output_size)

        def forward(self, x):
            return self.fc2(x)

    # 异步流水线训练函数
    def train_async_pipeline(
        rank, input_size, world_size, num_microbatches, microbatch_size, data_queue, result_queue
    ):
        # 设置分布式环境
        setup_distributed(rank, world_size)

        # 参数设置
        hidden_size = 32
        output_size = 2  # 二分类
        device = torch.device(f"cuda:{rank}")

        # 根据 rank 分配阶段
        if rank == 0:
            model = Stage1(input_size, hidden_size).to(device)
        elif rank == 1:
            model = Stage2(hidden_size, output_size).to(device)

        # 定义优化器
        optimizer = torch.optim.SGD(model.parameters(), lr=0.01)

        # 异步流水线处理
        model.train()
        for epoch in range(MOCK_EPOCH):
            optimizer.zero_grad()

            if rank == 0:
                intermediates = []
                # Stage 1: 处理输入并发送中间结果
                for i in range(num_microbatches):
                    try:
                        micro_data = data_queue.get(timeout=1)  # 从队列获取微批次数据
                        intermediate = model(micro_data.to(device))  # 前向计算
                        # print(f"isend intermediate: {intermediate} {intermediate.shape}")
                        dist.isend(tensor=intermediate, dst=1)  # 异步发送
                        intermediate = intermediate.detach().requires_grad_(
                            True
                        )  # 断开计算图，但保留梯度
                        intermediates.append(intermediate)  # 保存中间结果用于梯度传递
                    except queue.Empty:
                        print("no more data, break")
                        break

                # 接收并应用梯度
                for i in range(num_microbatches):
                    if i >= len(intermediates):
                        break
                    intermediate = intermediates[i]
                    grad = torch.zeros_like(intermediate).to(device)
                    dist.irecv(tensor=grad, src=1).wait()  # 异步接收梯度
                    # print(f"irecv intermediate.grad: {grad} {grad.shape}")
                    intermediate.backward(grad)

            elif rank == 1:
                total_loss = 0
                # Stage 2: 接收中间结果并计算输出
                for i in range(num_microbatches):
                    intermediate = torch.zeros(
                        microbatch_size, hidden_size, requires_grad=True, device=device
                    )
                    try:
                        # is_ok = dist.irecv(tensor=intermediate, src=0).wait()
                        is_ok = dist.irecv(tensor=intermediate, src=0).wait(
                            timeout=timedelta(seconds=3)
                        )
                    # except RuntimeError as e:
                    except RuntimeError:
                        # print(f"Error during irecv.wait: {e}")
                        is_ok = False

                    if not is_ok:
                        print("wait timeout, no data, next")
                        break
                    # print(f"irecv intermediate: {intermediate} {intermediate.shape}")
                    output = model(intermediate)

                    # 计算损失（假设标签为随机生成）
                    loss_fn = nn.CrossEntropyLoss()
                    micro_labels = torch.randint(0, output_size, (microbatch_size,)).to(device)
                    loss = loss_fn(output, micro_labels)
                    total_loss += loss.item()
                    loss.backward()

                    # 异步发送梯度回 Stage 1
                    assert intermediate.grad is not None, "intermediate.grad is None in Stage 2"
                    # print(f"isend intermediate.grad: {intermediate.grad} {intermediate.grad.shape}")
                    dist.isend(tensor=intermediate.grad, dst=0)
                total_loss = total_loss / num_microbatches
                print(f"Epoch {epoch+1} loss: {total_loss}\n")

            optimizer.step()
            torch.cuda.synchronize()  # 确保所有操作完成
            if rank == 0:
                print(f"Epoch {epoch+1} finished on Rank {rank}\n")

        cleanup()

    # 生成模拟数据
    batch_size = 4  # 微批次总和
    input_size = 16
    num_microbatches = 4  # 每个微批次大小
    data = torch.randn(batch_size, input_size)

    # 创建数据队列和结果队列
    data_queue = Queue()
    result_queue = Queue()

    # 将数据分片放入队列
    microbatch_size = batch_size // num_microbatches  # 假设 4 个微批次
    for _ in range(MOCK_EPOCH):
        for i in range(0, batch_size, microbatch_size):
            micro_data = data[i : i + microbatch_size]
            data_queue.put(micro_data)

    print(f"Running Asynchronous Pipeline Parallelism with {world_size} GPUs")

    # 启动进程
    processes = []
    for rank in range(world_size):
        p = Process(
            target=train_async_pipeline,
            args=(
                rank,
                input_size,
                world_size,
                num_microbatches,
                microbatch_size,
                data_queue,
                result_queue,
            ),
        )
        p.start()
        processes.append(p)

    for p in processes:
        p.join()



# https://docs.comfy.org/tutorials/flux/flux-2-dev
"""
📂 ComfyUI/
├── 📂 models/
│   ├── 📂 text_encoders/
│   │      └── mistral_3_small_flux2_bf16.safetensors
│   │      └── mistral_3_small_flux2_fp8.safetensors
│   ├── 📂 diffusion_models/
│   │      └── flux2_dev_fp8mixed.safetensors
│   └── 📂 vae/
│          └── flux2-vae.safetensors
"""

import os
import uuid
import json
import logging
from pathlib import Path
import subprocess

import modal

from . import model_dir, MODEL_VOL, comfyui_out_dir, COMFYUI_OUT_VOL, clear_comfyui_output_dir


# https://github.com/black-forest-labs/flux2
# https://huggingface.co/black-forest-labs/FLUX.2-dev
# https://huggingface.co/Comfy-Org/flux2-dev/tree/main/split_files
# modal run src/download_models.py --repo-ids "Comfy-Org/flux2-dev" --allow-patterns "*.safetensors"


def link_comfyui_dir():
    """
    ymlink the model to the right ComfyUI directory
    NOTE: need link ckpt file, don't to link models dir
    """
    clear_comfyui_output_dir()

    models_dir = f"{model_dir}/Comfy-Org/flux2-dev/split_files"
    for ckpt_file in Path(models_dir).glob("**/*.safetensors"):
        cmd = (
            f"ln -s {ckpt_file} /root/comfy/ComfyUI/models/{ckpt_file.parent.name}/{ckpt_file.name}"
        )
        print(cmd)
        subprocess.run(
            cmd,
            shell=True,
            check=True,
        )


def change_workflow_conf(file_path: Path, **kwargs) -> str:
    workflow_data = json.loads(file_path.read_text())

    # insert the prompt
    if kwargs.get("prompt"):
        workflow_data["6"]["inputs"]["text"] = kwargs.get("prompt")

    client_id = uuid.uuid4().hex
    workflow_data["9"]["inputs"]["filename_prefix"] = client_id

    # change output image size: width x height
    if kwargs.get("width"):
        workflow_data["47"]["inputs"]["width"] = kwargs.get("width")
        workflow_data["48"]["inputs"]["width"] = kwargs.get("width")
    if kwargs.get("height"):
        workflow_data["47"]["inputs"]["height"] = kwargs.get("height")
        workflow_data["48"]["inputs"]["height"] = kwargs.get("height")

    # ksample steps
    if kwargs.get("steps"):
        workflow_data["48"]["inputs"]["steps"] = kwargs.get("steps")

    images = kwargs.get("images")
    if images is not None and isinstance(images, list) and len(images) > 0:
        # now just support one image input
        if isinstance(images[0], bytes):
            input_image_filename = f"input_{client_id}.jpeg"
            input_image_path = f"{comfyui_out_dir}/{input_image_filename}"
            with open(input_image_path, "wb") as f:
                f.write(images[0])
            images[0] = input_image_path
            workflow_data["46"]["inputs"]["image"] = images[0]

    new_workflow_file = f"{client_id}.json"
    print(file_path, workflow_data)
    json.dump(workflow_data, Path(new_workflow_file).open("w", encoding="utf-8"))

    return client_id



"""
The code is base on https://github.com/Pointcept/Pointcept
"""

import os
import sys
import argparse
import multiprocessing as mp
from torch.nn.parallel import DistributedDataParallel


import utils.comm as comm
from utils.env import get_random_seed, set_seed
from utils.config import Config, DictAction


def create_ddp_model(model, *, fp16_compression=False, **kwargs):
    """
    Create a DistributedDataParallel model if there are >1 processes.
    Args:
        model: a torch.nn.Module
        fp16_compression: add fp16 compression hooks to the ddp object.
            See more at https://pytorch.org/docs/stable/ddp_comm_hooks.html#torch.distributed.algorithms.ddp_comm_hooks.default_hooks.fp16_compress_hook
        kwargs: other arguments of :module:`torch.nn.parallel.DistributedDataParallel`.
    """
    if comm.get_world_size() == 1:
        return model
    # kwargs['find_unused_parameters'] = True
    if "device_ids" not in kwargs:
        kwargs["device_ids"] = [comm.get_local_rank()]
        if "output_device" not in kwargs:
            kwargs["output_device"] = [comm.get_local_rank()]
    ddp = DistributedDataParallel(model, **kwargs)
    if fp16_compression:
        from torch.distributed.algorithms.ddp_comm_hooks import default as comm_hooks

        ddp.register_comm_hook(state=None, hook=comm_hooks.fp16_compress_hook)
    return ddp


def worker_init_fn(worker_id, num_workers, rank, seed):
    """Worker init func for dataloader.

    The seed of each worker equals to num_worker * rank + worker_id + user_seed

    Args:
        worker_id (int): Worker id.
        num_workers (int): Number of workers.
        rank (int): The rank of current process.
        seed (int): The random seed to use.
    """

    worker_seed = num_workers * rank + worker_id + seed
    set_seed(worker_seed)


def default_argument_parser(epilog=None):
    parser = argparse.ArgumentParser(
        epilog=epilog
        or f"""
    Examples:
    Run on single machine:
        $ {sys.argv[0]} --num-gpus 8 --config-file cfg.yaml
    Change some config options:
        $ {sys.argv[0]} --config-file cfg.yaml MODEL.WEIGHTS /path/to/weight.pth SOLVER.BASE_LR 0.001
    Run on multiple machines:
        (machine0)$ {sys.argv[0]} --machine-rank 0 --num-machines 2 --dist-url <URL> [--other-flags]
        (machine1)$ {sys.argv[0]} --machine-rank 1 --num-machines 2 --dist-url <URL> [--other-flags]
    """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--config-file", default="", metavar="FILE", help="path to config file"
    )
    parser.add_argument(
        "--num-gpus", type=int, default=1, help="number of gpus *per machine*"
    )
    parser.add_argument(
        "--num-machines", type=int, default=1, help="total number of machines"
    )
    parser.add_argument(
        "--machine-rank",
        type=int,
        default=0,
        help="the rank of this machine (unique per machine)",
    )
    # PyTorch still may leave orphan processes in multi-gpu training.
    # Therefore we use a deterministic way to obtain port,
    # so that users are aware of orphan processes by seeing the port occupied.
    # port = 2 ** 15 + 2 ** 14 + hash(os.getuid() if sys.platform != "win32" else 1) % 2 ** 14
    parser.add_argument(
        "--dist-url",
        # default="tcp://127.0.0.1:{}".format(port),
        default="auto",
        help="initialization URL for pytorch distributed backend. See "
        "https://pytorch.org/docs/stable/distributed.html for details.",
    )
    parser.add_argument(
        "--options", nargs="+", action=DictAction, help="custom options"
    )
    return parser


def default_config_parser(file_path, options):
    # config name protocol: dataset_name/model_name-exp_name
    if os.path.isfile(file_path):
        cfg = Config.fromfile(file_path)
    else:
        sep = file_path.find("-")
        cfg = Config.fromfile(os.path.join(file_path[:sep], file_path[sep + 1 :]))

    if options is not None:
        cfg.merge_from_dict(options)

    if cfg.seed is None:
        cfg.seed = get_random_seed()

    cfg.data.train.loop = cfg.epoch // cfg.eval_epoch

    os.makedirs(os.path.join(cfg.save_path, "model"), exist_ok=True)
    if not cfg.resume:
        cfg.dump(os.path.join(cfg.save_path, "config.py"))
    return cfg


def default_setup(cfg):
    # scalar by world size
    world_size = comm.get_world_size()
    cfg.num_worker = cfg.num_worker if cfg.num_worker is not None else mp.cpu_count()
    cfg.num_worker_per_gpu = cfg.num_worker // world_size
    assert cfg.batch_size % world_size == 0
    assert cfg.batch_size_val is None or cfg.batch_size_val % world_size == 0
    assert cfg.batch_size_test is None or cfg.batch_size_test % world_size == 0
    cfg.batch_size_per_gpu = cfg.batch_size // world_size
    cfg.batch_size_val_per_gpu = (
        cfg.batch_size_val // world_size if cfg.batch_size_val is not None else 1
    )
    cfg.batch_size_test_per_gpu = (
        cfg.batch_size_test // world_size if cfg.batch_size_test is not None else 1
    )
    # update data loop
    assert cfg.epoch % cfg.eval_epoch == 0
    # settle random seed
    rank = comm.get_rank()
    seed = None if cfg.seed is None else cfg.seed * cfg.num_worker_per_gpu + rank
    set_seed(seed)
    return cfg



import os
import subprocess
import asyncio

import modal

ASR_TAG = os.getenv("ASR_TAG", "whisper_trtllm_asr")

app = modal.App("trtllm-run-whisper-achatbot")
# https://github.com/NVIDIA/TensorRT-LLM/tree/v0.18.0/examples/whisper
# https://github.com/NVIDIA/TensorRT-LLM/blob/v0.20.0/examples/models/core/whisper/README.md
GIT_TAG_OR_HASH = os.getenv("GIT_TAG_OR_HASH", "v0.18.0")
WHISPER_DIRS = {
    "v0.18.0": "/TensorRT-LLM/examples/whisper",
    "v0.20.0": "/TensorRT-LLM/examples/models/core/whisper",
}

trtllm_image = (
    (
        # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
        modal.Image.from_registry(
            "nvidia/cuda:12.4.0-devel-ubuntu22.04",
            add_python="3.10",
        )
        .apt_install(
            "ffmpeg", "git", "clang", "git-lfs", "openmpi-bin", "libopenmpi-dev", "wget"
        )  # OpenMPI for distributed communication
        .pip_install(
            f"tensorrt-llm=={GIT_TAG_OR_HASH}",
            pre=True,
            extra_index_url="https://pypi.nvidia.com",
        )
        # https://github.com/flashinfer-ai/flashinfer/issues/738
        # .pip_install(
        #    "wheel",
        #    "setuptools==75.6.0",
        #    "packaging==23.2",
        #    "ninja==1.11.1.3",
        #    "build==1.2.2.post1",
        # )
        # .run_commands(
        #    "git clone https://github.com/flashinfer-ai/flashinfer.git --recursive",
        #    'cd /flashinfer && git checkout v0.2.5 && TORCH_CUDA_ARCH_LIST="7.5 8.0 8.9 9.0a" FLASHINFER_ENABLE_AOT=1 pip install --no-build-isolation --verbose --editable .',
        # )
        .run_commands(
            f"git clone https://github.com/NVIDIA/TensorRT-LLM.git -b {GIT_TAG_OR_HASH}",
            f"pip install -r {WHISPER_DIRS[GIT_TAG_OR_HASH]}/requirements.txt",
        )
    )
    .pip_install(
        f"achatbot=={os.getenv('ACHATBOT_VERSION', '0.0.24')}",
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    )
    .env(
        {
            "ACHATBOT_PKG": "1",
            "ASR_TAG": ASR_TAG,
        }
    )
)

TRT_MODEL_DIR = "/root/.achatbot/models"
trt_model_vol = modal.Volume.from_name("triton_trtllm_models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)


@app.function(
    gpu=os.getenv("IMAGE_GPU", "L4"),
    retries=0,
    image=trtllm_image,
    volumes={
        TRT_MODEL_DIR: trt_model_vol,
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
)
async def run(
    func: callable = None,
    **kwargs,
) -> None:
    import torch

    print("torch:", torch.__version__)
    print("cuda:", torch.version.cuda)
    print("_GLIBCXX_USE_CXX11_ABI", torch._C._GLIBCXX_USE_CXX11_ABI)

    subprocess.run("pip show achatbot", shell=True)
    subprocess.run("pip show tensorrt", shell=True)
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("which nvcc", shell=True)
    subprocess.run("nvcc --version", shell=True)

    print(f"{kwargs=}")
    if asyncio.iscoroutinefunction(func):
        await func(**kwargs)
    else:
        func(**kwargs)


async def transcribe(**kwargs):
    from achatbot.common.interface import IAsr
    from achatbot.modules.speech.asr import ASREnvInit
    from achatbot.common.session import Session
    from achatbot.common.types import SessionCtx
    from achatbot.common.logger import Logger

    Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

    asr: IAsr = ASREnvInit.initASREngine(os.getenv("ASR_TAG", "whisper_trtllm_asr"), **kwargs)
    audio_file = os.path.join(ASSETS_DIR, "asr_example_zh.wav")
    session = Session(**SessionCtx("test_client_id", 16000, 2).__dict__)
    asr.set_audio_data(audio_file)
    res = await asr.transcribe(session)
    print(res)


"""
# C++ runtime
ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --no-use-py-session
GIT_TAG_OR_HASH=v0.20.0 ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --no-use-py-session

ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --no-use-py-session --engine-dir trt_engines_float16_int8
GIT_TAG_OR_HASH=v0.20.0 ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --no-use-py-session --engine-dir trt_engines_float16_int8

GIT_TAG_OR_HASH=v0.20.0 ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --no-use-py-session --engine-dir trt_engines_float16_int4
GIT_TAG_OR_HASH=v0.20.0 ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --no-use-py-session --engine-dir trt_engines_float16_int4

# Python runtime
ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --use-py-session
ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --use-py-session --engine-dir trt_engines_float16_int8
ASR_TAG=whisper_trtllm_asr modal run src/llm/achatbot_audio.py --task transcribe --use-py-session --engine-dir trt_engines_float16_int4
"""


@app.local_entrypoint()
def main(
    task: str = "transcribe",
    engine_dir: str = "trt_engines_float16",
    max_batch_size: int = 1,  # need <= build decoder_model_config.max_batch_size
    use_py_session: bool = True,
):
    tasks = {
        "transcribe": transcribe,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    local_trt_build_dir = os.path.join(TRT_MODEL_DIR, "whisper", engine_dir)
    run.remote(
        func=tasks[task],
        engine_dir=local_trt_build_dir,
        max_batch_size=max_batch_size,
        use_py_session=use_py_session,
    )



import asyncio
import os
import logging
from typing import Any, Awaitable, Callable
import uuid

import unittest

from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.processors.logger import FrameLogger
from apipeline.frames.data_frames import TextFrame
from apipeline.frames.control_frames import EndFrame
from apipeline.frames.sys_frames import MetricsFrame

from src.common.utils.time import time_now_iso8601
from src.processors.aggregators.openai_llm_context import (
    OpenAILLMContext,
    OpenAIAssistantContextAggregator,
    OpenAIUserContextAggregator,
)
from src.processors.llm.base import LLMProcessor
from src.processors.llm.openai_llm_processor import OpenAIGroqLLMProcessor, OpenAILLMProcessor
from src.processors.vision.vision_processor import MockVisionProcessor, VisionProcessor
from src.common.session import Session
from src.common.types import SessionCtx
from src.core.llm import LLMEnvInit
from src.common.logger import Logger
from src.types.frames.data_frames import LLMMessagesFrame, TranscriptionFrame
from src.types.frames.control_frames import UserStartedSpeakingFrame, UserStoppedSpeakingFrame

from dotenv import load_dotenv

load_dotenv()


"""
python -m unittest test.integration.processors.test_openai_llm_processor.TestProcessor.test_run


MODEL=llama-3.1-70b-versatile \
    python -m unittest test.integration.processors.test_openai_llm_processor.TestProcessor.test_run

BASE_URL=https://api.together.xyz/v1 MODEL=meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo \
    python -m unittest test.integration.processors.test_openai_llm_processor.TestProcessor.test_run

BASE_URL=http://127.0.0.1:11434/v1 MODEL=qwen3:0.6b \
    python -m unittest test.integration.processors.test_openai_llm_processor.TestProcessor.test_run

BASE_URL=https://open.bigmodel.cn/api/paas/v4 MODEL=glm-4.5-flash \
    python -m unittest test.integration.processors.test_openai_llm_processor.TestProcessor.test_run
"""


class TestProcessor(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False)
        cls.max_function_call_cn = int(os.environ.get("MAX_FUNCTION_CALL_CN", "3"))
        cls.messages = [
            {
                "role": "system",
                # "content": "You are a weather assistant. Use the get_weather function to retrieve weather information for a given location."
                "content": "You are a helpful assistant who converses with a user and answers questions. Respond concisely to general questions.  Your response will be turned into speech so use only simple words and punctuation.\n  You have access to two tools: get_weather and describe_image.  You can respond to questions about the weather using the get_weather tool.\n  You can answer questions about the user's video stream using the describe_image tool.\n Some examples of phrases that indicate you should use the describe_image tool are: \n - What do you see?  \n - What's in the video? \n - Can you describe the video?\n - Tell me about what you see.\n  - Tell me something interesting about what you see.\n  - What's happening in the video?\n  If you need to use a tool, simply use the tool. Do not tell the user the tool you are using. Be brief and concise.\n Please communicate in Chinese",
            }
        ]
        cls.tools = [
            {
                "function": {
                    "description": "Get the current weather in a given location",
                    "name": "get_weather",
                    "parameters": {
                        "properties": {
                            "location": {
                                "description": "The city and state, e.g. San Francisco, CA",
                                "type": "string",
                            }
                        },
                        "required": ["location"],
                        "type": "object",
                    },
                },
                "type": "function",
            },
        ]

    @classmethod
    def tearDownClass(cls):
        pass

    async def get_weather(
        self,
        function_name: str,
        tool_call_id: str,
        arguments: Any,
        llm: LLMProcessor,
        context: OpenAILLMContext,
        result_callback: Callable[[Any], Awaitable[None]],
    ):
        location = arguments["location"]
        logging.info(
            f"function_name:{function_name}, tool_call_id:{tool_call_id},"
            f"arguments:{arguments}, llm:{llm}, context:{context}"
        )
        # just a mock response
        # add result to assistant context
        self.get_weather_call_cn += 1
        if self.max_function_call_cn > self.get_weather_call_cn:
            await result_callback(f"The weather in {location} is currently 72 degrees and sunny.")
        else:
            self.get_weather_call_cn = 0

    def get_vision_llm_processor(self) -> VisionProcessor:
        if self.vision_tag == "mock_vision_processor":
            return MockVisionProcessor(self.mock_text)

        session = Session(**SessionCtx(str(uuid.uuid4())).__dict__)
        llm = LLMEnvInit.initLLMEngine()
        return VisionProcessor(llm=llm, session=session)

    def get_openai_llm_processor(self) -> LLMProcessor:
        # default use openai llm processor
        api_key = os.environ.get("OPENAI_API_KEY")
        base_url = os.environ.get("BASE_URL", "https://api.groq.com/openai/v1")
        model = os.environ.get("MODEL", "llama3-groq-70b-8192-tool-use-preview")
        if "groq" in base_url:
            # https://console.groq.com/docs/models
            api_key = os.environ.get("GROQ_API_KEY")
            llm_processor = OpenAIGroqLLMProcessor(
                model=model,
                base_url=base_url,
                api_key=api_key,
            )
            return llm_processor
        elif "together" in base_url:
            # https://docs.together.ai/docs/chat-models
            api_key = os.environ.get("TOGETHER_API_KEY")
        elif "bigmodel" in base_url:
            # https://docs.bigmodel.cn/cn/guide/models/
            api_key = os.environ.get("ZHIPU_API_KEY")
        elif "local" in base_url or "127.0.0.1" in base_url:
            api_key = "ollama"
        llm_processor = OpenAILLMProcessor(
            model=model,
            base_url=base_url,
            api_key=api_key,
        )
        return llm_processor

    async def asyncSetUp(self):
        self.llm_context = OpenAILLMContext()
        self.llm_context.set_messages(self.messages)
        self.llm_context.set_tools(self.tools)
        llm_user_ctx_aggr = OpenAIUserContextAggregator(self.llm_context)
        llm_assistant_ctx_aggr = OpenAIAssistantContextAggregator(llm_user_ctx_aggr)

        llm_processor = self.get_openai_llm_processor()
        llm_processor.register_function("get_weather", self.get_weather)
        self.get_weather_call_cn = 0

        pipeline = Pipeline(
            [
                llm_user_ctx_aggr,
                llm_processor,
                llm_assistant_ctx_aggr,
                FrameLogger(include_frame_types=[MetricsFrame]),
            ]
        )
        self.task = PipelineTask(
            pipeline,
            params=PipelineParams(enable_metrics=True),
        )

    async def asyncTearDown(self):
        pass

    async def test_run(self):
        runner = PipelineRunner()
        await self.task.queue_frames(
            [
                UserStartedSpeakingFrame(),
                TranscriptionFrame(
                    "What's the weather like in New York today? ", "", time_now_iso8601(), "en"
                ),
                UserStoppedSpeakingFrame(),
                EndFrame(),
            ]
        )
        await runner.run(self.task)
        msgs = self.llm_context.get_messages()
        print(msgs)
        print("get_weather_call_cn", self.get_weather_call_cn)
        self.assertLess(self.get_weather_call_cn, self.max_function_call_cn)
        self.assertEqual(len(msgs) % 2, 1)
        if len(msgs) == 3:
            self.assertEqual(msgs[0]["role"], "system")
            self.assertEqual(msgs[1]["role"], "user")
            self.assertEqual(msgs[2]["role"], "assistant")
        if len(msgs) == 5:
            self.assertEqual(msgs[0]["role"], "system")
            self.assertEqual(msgs[1]["role"], "user")
            self.assertEqual(msgs[2]["role"], "assistant")
            self.assertEqual(msgs[3]["role"], "tool")
            self.assertEqual(msgs[4]["role"], "assistant")



import datetime
import time

from collections.abc import AsyncIterator, Callable, Iterable
from typing import Any

from a2a.client import (
    Client,
    ClientCallInterceptor,
    ClientEvent,
    ClientFactory,
    Consumer,
)
from a2a.client.client_factory import TransportProducer
from a2a.client.middleware import ClientCallContext
from a2a.extensions.common import HTTP_EXTENSION_HEADER, find_extension_by_uri
from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.events.event_queue import EventQueue
from a2a.types import (
    AgentCard,
    AgentExtension,
    Artifact,
    GetTaskPushNotificationConfigParams,
    Message,
    Role,
    SendMessageRequest,
    SendStreamingMessageRequest,
    Task,
    TaskArtifactUpdateEvent,
    TaskIdParams,
    TaskPushNotificationConfig,
    TaskQueryParams,
    TaskStatusUpdateEvent,
)


_CORE_PATH = "github.com/a2aproject/a2a-samples/extensions/timestamp/v1"
URI = f"https://{_CORE_PATH}"
TIMESTAMP_FIELD = f"{_CORE_PATH}/timestamp"


class TimestampExtension:
    """An implementation of the Timestamp extension.

    This extension implementation illustrates several ways for an extension to
    provide functionality to agent developers. In general, the support methods
    range from totally hands off, where all responsibility for using the
    extension correctly is left to the developer, to totally hands-on, where
    the developer sets up strategic decorators for core classes which then
    manage implementing the extension logic. Each of the methods have comments
    indicating the level of support they provide.
    """

    def __init__(self, now_fn: Callable[[], float] | None = None):
        self._now_fn = now_fn or time.time

    # Option 1 for adding to a card: let the developer do it themselves.
    def agent_extension(self) -> AgentExtension:
        """Get the AgentExtension representing this extension."""
        return AgentExtension(
            uri=URI,
            description="Adds timestamps to messages and artifacts.",
        )

    # Option 2 for adding to a card: do it for them.
    def add_to_card(self, card: AgentCard) -> AgentCard:
        """Add this extension to an AgentCard."""
        if not (exts := card.capabilities.extensions):
            exts = card.capabilities.extensions = []
        exts.append(self.agent_extension())
        return card

    def is_supported(self, card: AgentCard | None) -> bool:
        """Returns whether this extension is supported by the AgentCard."""
        if card:
            return find_extension_by_uri(card, URI) is not None
        return False

    def activate(self, context: RequestContext) -> bool:
        """Possibly activate this extension, depending on the request context.

        The extension is considered active if the caller indicated it in an
        X-A2A-Extensions header.
        """
        if URI in context.requested_extensions:
            context.add_activated_extension(URI)
            return True
        return False

    # Option 1 for adding to a message: self-serve.
    def add_timestamp(self, o: Message | Artifact) -> None:
        """Add a timestamp to a message or artifact."""
        # Respect existing timestamps.
        if self.has_timestamp(o):
            return
        if o.metadata is None:
            o.metadata = {}
        now = self._now_fn()
        dt = datetime.datetime.fromtimestamp(now, datetime.UTC)
        o.metadata[TIMESTAMP_FIELD] = dt.isoformat()

    # Option 2: assisted, but still self-serve
    def add_if_activated(self, o: Message | Artifact, context: RequestContext) -> None:
        """Add a timestamp to a message or artifact if the extension is active."""
        if self.activate(context):
            self.add_timestamp(o)

    # Option 3 for servers: timestamp an event.
    def timestamp_event(
        self,
        event: Message | Task | TaskStatusUpdateEvent | TaskArtifactUpdateEvent,
    ) -> None:
        """Add a timestamp to a server-side event."""
        for o in self._get_messages_in_event(event):
            self.add_timestamp(o)

    # Option 4: helper class
    def get_timestamper(self, context: RequestContext) -> "MessageTimestamper":
        """Returns a helper class for adding timestamps to messages/artifacts.

        This detects whether the extension should be activated based on the
        current RequestContext. If not, timestamps are not added.
        """
        active = self.activate(context)
        return MessageTimestamper(active, self)

    def get_timestamp(self, o: Message | Artifact) -> datetime.datetime | None:
        """Get a timestamp from a message or artifact."""
        if o.metadata and (ts := o.metadata.get(TIMESTAMP_FIELD)):
            return datetime.datetime.fromisoformat(ts)
        return None

    def has_timestamp(self, o: Message | Artifact) -> bool:
        """Returns whether a message or artifact has a timestamp."""
        if o.metadata:
            return TIMESTAMP_FIELD in o.metadata
        return False

    # Option 5: Fully managed via a decorator. This is the most complicated, but
    # easiest for a developer to use.
    def wrap_executor(self, executor: AgentExecutor) -> AgentExecutor:
        """Wrap an executor in a decorator that automatically adds timestamps to messages and artifacts."""
        return _TimestampingAgentExecutor(executor, self)

    def request_activation_http(self, http_kwargs: dict[str, Any]) -> dict[str, Any]:
        """Update an http_kwargs to request activation of this extension."""
        if not (headers := http_kwargs.get("headers")):
            headers = http_kwargs["headers"] = {}
        header_val = URI
        if headers.get(HTTP_EXTENSION_HEADER):
            header_val = headers[HTTP_EXTENSION_HEADER] + ", " + URI
        headers[HTTP_EXTENSION_HEADER] = header_val
        return http_kwargs

    # Option 2 for clients: timestamp your JSON RPC payloads.
    # Option 1 is to self-serve add the timestamp to your message.
    def timestamp_request_message(
        self, request: SendMessageRequest | SendStreamingMessageRequest
    ) -> None:
        """Add a timestamp to an outgoing request."""
        self.add_timestamp(request.params.message)

    # Option 3 for clients: use a client interceptor.
    def client_interceptor(self) -> ClientCallInterceptor:
        """Get a client interceptor that activates this extension."""
        return _TimestampingClientInterceptor(self)

    # Option 4 for clients: wrap the client itself.
    def wrap_client(self, client: Client) -> Client:
        """Returns a Client that ensures all outgoing messages have timestamps."""
        return _TimestampingClient(client, self)

    # Option 5 for clients: an extension-aware client factory.
    def wrap_client_factory(self, factory: ClientFactory) -> ClientFactory:
        """Returns a ClientFactory that handles this extension."""
        return _TimestampClientFactory(factory, self)

    def _get_messages_in_event(
        self,
        event: Message | Task | TaskStatusUpdateEvent | TaskArtifactUpdateEvent,
    ) -> Iterable[Message | Artifact]:
        if isinstance(event, TaskStatusUpdateEvent) and event.status.message:
            return [event.status.message]
        if isinstance(event, TaskArtifactUpdateEvent):
            return [event.artifact]
        if isinstance(event, Message):
            return [event]
        if isinstance(event, Task):
            return self._get_artifacts_and_messages_in_task(event)
        return []

    def _get_artifacts_and_messages_in_task(self, t: Task) -> Iterable[Message | Artifact]:
        if t.artifacts:
            yield from t.artifacts
        if t.history:
            yield from (m for m in t.history if m.role == Role.agent)
        if t.status.message:
            yield t.status.message


class MessageTimestamper:
    """Helper to add compliant timestamps to messages and artifacts.

    Timestamps are only added if the extension is activated."""

    def __init__(self, active: bool, ext: TimestampExtension):
        self._active = active
        self._ext = ext

    def timestamp(self, o: Message | Artifact) -> None:
        """Add a timestamp to a message or artifact, if active."""
        if self._active:
            self._ext.add_timestamp(o)


class _TimestampingAgentExecutor(AgentExecutor):
    def __init__(self, delegate: AgentExecutor, ext: TimestampExtension):
        self._delegate = delegate
        self._ext = ext

    async def execute(self, context: RequestContext, event_queue: EventQueue) -> None:
        # Wrap the EventQueue so that all outgoing messages/status updates have
        # timestamps.
        return await self._delegate.execute(context, self._maybe_wrap_queue(context, event_queue))

    def _maybe_wrap_queue(self, context: RequestContext, queue: EventQueue) -> EventQueue:
        if self._ext.activate(context):
            return _TimestampingEventQueue(queue, self._ext)
        return queue

    async def cancel(self, context: RequestContext, event_queue: EventQueue) -> None:
        return await self._delegate.cancel(context, event_queue)


class _TimestampingEventQueue(EventQueue):
    """An EventQueue decorator that adds timestamps to all events."""

    def __init__(self, delegate: EventQueue, ext: TimestampExtension):
        self._delegate = delegate
        self._ext = ext

    async def enqueue_event(
        self,
        event: Message | Task | TaskStatusUpdateEvent | TaskArtifactUpdateEvent,
    ) -> None:
        # If we're here, we're activated. Timestamp everything.
        self._ext.timestamp_event(event)
        return await self._delegate.enqueue_event(event)

    # Finish out all delegate methods.

    async def dequeue_event(
        self, no_wait: bool = False
    ) -> Message | Task | TaskStatusUpdateEvent | TaskArtifactUpdateEvent:
        return await self._delegate.dequeue_event(no_wait)

    async def close(self) -> None:
        return await self._delegate.close()

    def tap(self) -> EventQueue:
        return self._delegate.tap()

    def is_closed(self) -> bool:
        return self._delegate.is_closed()

    def task_done(self) -> None:
        return self._delegate.task_done()


_MESSAGING_METHODS = {"message/send", "message/stream"}


class _TimestampClientFactory(ClientFactory):
    """A ClientFactory decorator to aid in adding timestamps.

    This factory determines if agents support the timestamp extension, and, if
    so, ensures that outgoing messages have timestamps.
    """

    def __init__(self, delegate: ClientFactory, ext: TimestampExtension):
        self._delegate = delegate
        self._ext = ext

    def register(self, label: str, generator: TransportProducer) -> None:
        self._delegate.register(label, generator)

    def create(
        self,
        card: AgentCard,
        consumers: list[Consumer] | None = None,
        interceptors: list[ClientCallInterceptor] | None = None,
    ) -> Client:
        interceptors = interceptors or []
        interceptors.append(self._ext.client_interceptor())
        return self._delegate.create(card, consumers, interceptors)


class _TimestampingClient(Client):
    """A Client decorator that adds timestamps to outgoing messages."""

    def __init__(self, delegate: Client, ext: TimestampExtension):
        self._delegate = delegate
        self._ext = ext

    async def send_message(
        self,
        request: Message,
        *,
        context: ClientCallContext | None = None,
    ) -> AsyncIterator[ClientEvent | Message]:
        self._ext.add_timestamp(request)
        async for e in self._delegate.send_message(request, context=context):
            yield e

    async def get_task(
        self,
        request: TaskQueryParams,
        *,
        context: ClientCallContext | None = None,
    ) -> Task:
        return await self._delegate.get_task(request, context=context)

    async def cancel_task(
        self, request: TaskIdParams, *, context: ClientCallContext | None = None
    ) -> Task:
        return await self._delegate.cancel_task(request, context=context)

    async def set_task_callback(
        self,
        request: TaskPushNotificationConfig,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        return await self._delegate.set_task_callback(request, context=context)

    async def get_task_callback(
        self,
        request: GetTaskPushNotificationConfigParams,
        *,
        context: ClientCallContext | None = None,
    ) -> TaskPushNotificationConfig:
        return await self._delegate.get_task_callback(request, context=context)

    async def resubscribe(
        self, request: TaskIdParams, *, context: ClientCallContext | None = None
    ) -> AsyncIterator[ClientEvent]:
        async for e in self._delegate.resubscribe(request, context=context):
            yield e

    async def get_card(self, *, context: ClientCallContext | None = None) -> AgentCard:
        return await self._delegate.get_card(context=context)


class _TimestampingClientInterceptor(ClientCallInterceptor):
    """A client interceptor that adds timestamps to outgoing messages."""

    def __init__(self, ext: TimestampExtension):
        self._ext = ext

    async def intercept(
        self,
        method_name: str,
        request_payload: dict[str, Any],
        http_kwargs: dict[str, Any],
        agent_card: AgentCard | None,
        context: ClientCallContext | None,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        if not self._ext.is_supported(agent_card) or method_name not in _MESSAGING_METHODS:
            return (request_payload, http_kwargs)
        body: SendMessageRequest | SendStreamingMessageRequest
        if method_name == "message/send":
            body = SendMessageRequest.model_validate(request_payload)
        else:
            body = SendStreamingMessageRequest.model_validate(request_payload)
        self._ext.timestamp_request_message(body)
        # Request that we activate the extension, and timestamp the message.
        return (
            body.model_dump(),
            self._ext.request_activation_http(http_kwargs),
        )



#!/usr/bin/env python3
"""
Simple MCP client example with OAuth authentication support.

This client connects to an MCP server using streamable HTTP transport with OAuth.

from:
- https://github.com/modelcontextprotocol/python-sdk/blob/main/examples/clients/simple-auth-client/README.md
- https://modelcontextprotocol.io/specification/2025-03-26/basic/authorization
    - 1. OAuth Grant Types
        - Generate code_verifier and code_challenge
        - User-Agent (Browser) to get Authorization Code from /authorizes api with Authorization Request
        - auth code + code_verifier get Access Token (+ Refresh Token)
        - MCP Request with Access Token
    - 2. Authorization flow
        - GET /.well-known/oauth-authorization-server to get OAuth Server Metadata
        - POST /register to register client
        - Generate PKCE Parameters include code_verifier and code_challenge
        - User-Agent (Browser) to get Authorization Code from /authorizes api with Authorization Request
        - auth code + code_verifier get Access Token (+ Refresh Token)
        - MCP Request with Access Token
    - 3. Third-Party Authorization Flow
        - MCP client initiates standard OAuth flow with MCP server
        - MCP server redirects user to third-party authorization server
        - User authorizes with third-party server
        - Third-party server redirects back to MCP server with authorization code
        - MCP server exchanges code for third-party access token
        - MCP server generates its own access token bound to the third-party session
        - MCP server completes original OAuth flow with MCP client


this code use 2.Authorization flow
"""

import asyncio
import os
import threading
import time
import webbrowser
from datetime import timedelta
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Any
from urllib.parse import parse_qs, urlparse

from mcp.client.auth import OAuthClientProvider, TokenStorage
from mcp.client.session import ClientSession
from mcp.client.sse import sse_client
from mcp.client.streamable_http import streamablehttp_client
from mcp.shared.auth import OAuthClientInformationFull, OAuthClientMetadata, OAuthToken

from mcp_server.auth.token_storage import InMemoryTokenStorage


class CallbackHandler(BaseHTTPRequestHandler):
    """Simple HTTP handler to capture OAuth callback."""

    def __init__(self, request, client_address, server, callback_data):
        """Initialize with callback data storage."""
        self.callback_data = callback_data
        super().__init__(request, client_address, server)

    def do_GET(self):
        """Handle GET request from OAuth redirect."""
        parsed = urlparse(self.path)
        query_params = parse_qs(parsed.query)

        if "code" in query_params:
            self.callback_data["authorization_code"] = query_params["code"][0]
            self.callback_data["state"] = query_params.get("state", [None])[0]
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"""
            <html>
            <body>
                <h1>Authorization Successful!</h1>
                <p>You can close this window and return to the terminal.</p>
                <script>setTimeout(() => window.close(), 2000);</script>
            </body>
            </html>
            """)
        elif "error" in query_params:
            self.callback_data["error"] = query_params["error"][0]
            self.send_response(400)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(
                f"""
            <html>
            <body>
                <h1>Authorization Failed</h1>
                <p>Error: {query_params['error'][0]}</p>
                <p>You can close this window and return to the terminal.</p>
            </body>
            </html>
            """.encode()
            )
        else:
            self.send_response(404)
            self.end_headers()

    def log_message(self, format, *args):
        """Suppress default logging."""
        pass


class CallbackServer:
    """Simple server to handle OAuth callbacks."""

    def __init__(self, port=3000):
        self.port = port
        self.server = None
        self.thread = None
        self.callback_data = {"authorization_code": None, "state": None, "error": None}

    def _create_handler_with_data(self):
        """Create a handler class with access to callback data."""
        callback_data = self.callback_data

        class DataCallbackHandler(CallbackHandler):
            def __init__(self, request, client_address, server):
                super().__init__(request, client_address, server, callback_data)

        return DataCallbackHandler

    def start(self):
        """Start the callback server in a background thread."""
        handler_class = self._create_handler_with_data()
        self.server = HTTPServer(("localhost", self.port), handler_class)
        self.thread = threading.Thread(target=self.server.serve_forever, daemon=True)
        self.thread.start()
        print(f"🖥️  Started callback server on http://localhost:{self.port}")

    def stop(self):
        """Stop the callback server."""
        if self.server:
            self.server.shutdown()
            self.server.server_close()
        if self.thread:
            self.thread.join(timeout=1)

    def wait_for_callback(self, timeout=300):
        """Wait for OAuth callback with timeout."""
        start_time = time.time()
        while time.time() - start_time < timeout:
            if self.callback_data["authorization_code"]:
                return self.callback_data["authorization_code"]
            elif self.callback_data["error"]:
                raise Exception(f"OAuth error: {self.callback_data['error']}")
            time.sleep(0.1)
        raise Exception("Timeout waiting for OAuth callback")

    def get_state(self):
        """Get the received state parameter."""
        return self.callback_data["state"]


class SimpleAuthClient:
    """Simple MCP client with auth support."""

    def __init__(self, server_url: str, transport_type: str = "streamable_http"):
        self.server_url = server_url
        self.transport_type = transport_type
        self.session: ClientSession | None = None

    async def connect(self):
        """Connect to the MCP server."""
        print(f"🔗 Attempting to connect to {self.server_url}...")

        try:
            # Set up callback server
            callback_server = CallbackServer(port=3000)
            callback_server.start()

            async def callback_handler() -> tuple[str, str | None]:
                """Wait for OAuth callback and return auth code and state."""
                print("⏳ Waiting for authorization callback...")
                try:
                    auth_code = callback_server.wait_for_callback(timeout=300)
                    return auth_code, callback_server.get_state()
                finally:
                    callback_server.stop()

            client_metadata_dict = {
                "client_name": "Simple Auth Client",
                "redirect_uris": ["http://localhost:3000/callback"],
                "grant_types": ["authorization_code", "refresh_token"],
                "response_types": ["code"],
                "token_endpoint_auth_method": "client_secret_post",
            }

            async def _default_redirect_handler(authorization_url: str) -> None:
                """Default redirect handler that opens the URL in a browser."""
                print(f"Opening browser for authorization: {authorization_url}")
                webbrowser.open(authorization_url)

            # Create OAuth authentication handler using the new interface
            oauth_auth = OAuthClientProvider(
                server_url=self.server_url.replace("/mcp", ""),
                client_metadata=OAuthClientMetadata.model_validate(client_metadata_dict),
                storage=InMemoryTokenStorage(),
                redirect_handler=_default_redirect_handler,
                callback_handler=callback_handler,
            )

            # Create transport with auth handler based on transport type
            if self.transport_type == "sse":
                print("📡 Opening SSE transport connection with auth...")
                async with sse_client(
                    url=self.server_url,
                    auth=oauth_auth,
                    timeout=60,
                ) as (read_stream, write_stream):
                    await self._run_session(read_stream, write_stream, None)
            else:
                print("📡 Opening StreamableHTTP transport connection with auth...")
                async with streamablehttp_client(
                    url=self.server_url,
                    auth=oauth_auth,
                    timeout=timedelta(seconds=60),
                ) as (read_stream, write_stream, get_session_id):
                    await self._run_session(read_stream, write_stream, get_session_id)

        except Exception as e:
            print(f"❌ Failed to connect: {e}")
            import traceback

            traceback.print_exc()

    async def _run_session(self, read_stream, write_stream, get_session_id):
        """Run the MCP session with the given streams."""
        print("🤝 Initializing MCP session...")
        async with ClientSession(read_stream, write_stream) as session:
            self.session = session
            print("⚡ Starting session initialization...")
            await session.initialize()
            print("✨ Session initialization complete!")

            print(f"\n✅ Connected to MCP server at {self.server_url}")
            if get_session_id:
                session_id = get_session_id()
                if session_id:
                    print(f"Session ID: {session_id}")

            # Run interactive loop
            await self.interactive_loop()

    async def list_tools(self):
        """List available tools from the server."""
        if not self.session:
            print("❌ Not connected to server")
            return

        try:
            result = await self.session.list_tools()
            if hasattr(result, "tools") and result.tools:
                print("\n📋 Available tools:")
                for i, tool in enumerate(result.tools, 1):
                    print(f"{i}. {tool.name}")
                    if tool.description:
                        print(f"   Description: {tool.description}")
                    print()
            else:
                print("No tools available")
        except Exception as e:
            print(f"❌ Failed to list tools: {e}")

    async def call_tool(self, tool_name: str, arguments: dict[str, Any] | None = None):
        """Call a specific tool."""
        if not self.session:
            print("❌ Not connected to server")
            return

        try:
            result = await self.session.call_tool(tool_name, arguments or {})
            print(f"\n🔧 Tool '{tool_name}' result:")
            if hasattr(result, "content"):
                for content in result.content:
                    if content.type == "text":
                        print(content.text)
                    else:
                        print(content)
            else:
                print(result)
        except Exception as e:
            print(f"❌ Failed to call tool '{tool_name}': {e}")

    async def interactive_loop(self):
        """Run interactive command loop."""
        print("\n🎯 Interactive MCP Client")
        print("Commands:")
        print("  list - List available tools")
        print("  call <tool_name> [args] - Call a tool")
        print("  quit - Exit the client")
        print()

        while True:
            try:
                command = input("mcp> ").strip()

                if not command:
                    continue

                if command == "quit":
                    break

                elif command == "list":
                    await self.list_tools()

                elif command.startswith("call "):
                    parts = command.split(maxsplit=2)
                    tool_name = parts[1] if len(parts) > 1 else ""

                    if not tool_name:
                        print("❌ Please specify a tool name")
                        continue

                    # Parse arguments (simple JSON-like format)
                    arguments = {}
                    if len(parts) > 2:
                        import json

                        try:
                            arguments = json.loads(parts[2])
                        except json.JSONDecodeError:
                            print("❌ Invalid arguments format (expected JSON)")
                            continue

                    await self.call_tool(tool_name, arguments)

                else:
                    print("❌ Unknown command. Try 'list', 'call <tool_name>', or 'quit'")

            except KeyboardInterrupt:
                print("\n\n👋 Goodbye!")
                break
            except EOFError:
                break


async def main():
    """Main entry point."""
    # Default server URL - can be overridden with environment variable
    # Most MCP streamable HTTP servers use /mcp as the endpoint
    server_url = os.getenv("MCP_SERVER_PORT", 8000)
    transport_type = os.getenv("MCP_TRANSPORT_TYPE", "streamable_http")
    server_url = (
        f"http://localhost:{server_url}/mcp"
        if transport_type == "streamable_http"
        else f"http://localhost:{server_url}/sse"
    )

    print("🚀 Simple MCP Auth Client")
    print(f"Connecting to: {server_url}")
    print(f"Transport type: {transport_type}")

    # Start connection flow - OAuth will be handled automatically
    client = SimpleAuthClient(server_url, transport_type)
    await client.connect()


def cli():
    """CLI entry point for uv script."""
    asyncio.run(main())


if __name__ == "__main__":
    cli()



import os
import sys
import logging
import time
from typing import AsyncGenerator

import numpy as np

try:
    cur_dir = os.path.dirname(__file__)
    if bool(os.getenv("ACHATBOT_PKG", "")):
        sys.path.insert(0, os.path.join(cur_dir, "../../../HiggsAudio"))
    else:
        sys.path.insert(0, os.path.join(cur_dir, "../../../../deps/HiggsAudio"))
    from achatbot.HiggsAudio.boson_multimodal.serve.serve_engine import HiggsAudioServeEngine
    from achatbot.HiggsAudio.boson_multimodal.data_types import ChatMLSample, Message, AudioContent
    from achatbot.HiggsAudio.boson_multimodal.model.higgs_audio.utils import revert_delay_pattern

except ModuleNotFoundError as e:
    logging.error(f"Exception: {e}")
    logging.error(
        "In order to use Higgs-Audio, you need to `pip install achatbot[llm_transformers_manual_speech_higgs]`,"
    )
    raise Exception(f"Missing module: {e}")

from achatbot.common.utils.helper import get_device
from achatbot.common.session import Session
from achatbot.types.llm.transformers import TransformersSpeechLMArgs
from .base import TransformersBaseLLM


class TransformersManualSpeechHiggs(TransformersBaseLLM):
    """
    TTS: text + ref audio -> llama3 -> text + vq code tokens
    """

    TAG = "llm_transformers_manual_speech_higgs"
    DEFAULT_SYS_PROMPT = "Generate audio following instruction.\n\n<|scene_desc_start|>\nAudio is recorded from a quiet room.\n<|scene_desc_end|>"

    def __init__(self, **args):
        audio_tokenizer_path = args.pop("audio_tokenizer_path", None)
        self.args = TransformersSpeechLMArgs(**args)
        self.args.lm_device = self.args.lm_device or get_device()
        logging.info("TransformersLMArgs: %s", self.args)

        self.serve_engine = HiggsAudioServeEngine(
            self.args.lm_model_name_or_path,
            audio_tokenizer_path,
            device=self.args.lm_device,
            torch_dtype=self.args.lm_torch_dtype,
        )

        self.warmup()

    def warmup(self):
        if self.args.warmup_steps <= 0 or not self.args.warmup_prompt:
            logging.info("no warmup!")
            return

        messages = [
            Message(
                role="system",
                content=self.DEFAULT_SYS_PROMPT,
            ),
            Message(
                role="user",
                content=self.args.warmup_prompt,
            ),
        ]
        for step in range(self.args.warmup_steps):
            start_time = time.perf_counter()
            _ = self.serve_engine.generate(
                chat_ml_sample=ChatMLSample(messages=messages),
                max_new_tokens=self.args.lm_gen_max_new_tokens,
                temperature=self.args.lm_gen_temperature,
                top_p=self.args.lm_gen_top_p,
                top_k=self.args.lm_gen_top_k,
                stop_strings=["<|end_of_text|>", "<|eot_id|>"],
            )
            cost = time.perf_counter() - start_time
            logging.info(f"step {step} | warmup cost: {cost:.3f} s")

    async def async_generate(
        self, session: Session, **kwargs
    ) -> AsyncGenerator[str | dict | np.ndarray, None]:
        messages = session.ctx.state.get("messages", [])
        assert isinstance(messages, list) and len(messages) > 0, (
            "messages should be a list of Message objects"
        )
        streamer = self.serve_engine.generate_delta_stream(
            chat_ml_sample=ChatMLSample(messages=messages),
            max_new_tokens=kwargs.get("max_new_tokens", self.args.lm_gen_max_new_tokens),
            temperature=kwargs.get("temperature", self.args.lm_gen_temperature),
            top_p=kwargs.get("top_p", self.args.lm_gen_top_p),
            top_k=kwargs.get("top_k", self.args.lm_gen_top_k),
            stop_strings=kwargs.get("stop_strings", ["<|end_of_text|>", "<|eot_id|>"]),
        )

        times = []
        start_time = time.perf_counter()
        async for delta in streamer:
            times.append(time.perf_counter() - start_time)
            yield {
                "text": delta.text,
                "text_tokens": delta.text_tokens,
                "audio_vq_tokens": delta.audio_tokens,
                "finish_reason": delta.finish_reason,
            }
            start_time = time.perf_counter()

        if len(times) > 0:
            logging.info(f"gen TTFT time: {times[0]} s | total: {sum(times)} s")
        else:
            logging.warning("no generate stream")



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

import sys

from enum import Enum

if sys.version_info < (3, 11):

    class StrEnum(str, Enum):
        def __new__(cls, value):
            obj = str.__new__(cls, value)
            obj._value_ = value
            return obj
else:
    from enum import StrEnum


class Language(StrEnum):
    BG = "bg"  # Bulgarian
    CA = "ca"  # Catalan
    ZH = "zh"  # Chinese simplified
    ZH_CN = "zh-CN"  # Chinese simplified
    ZH_TW = "zh-TW"  # Chinese traditional
    ZN = "zn"  # Chinese and English
    CS = "cs"  # Czech
    DA = "da"  # Danish
    NL = "nl"  # Dutch
    EN = "en"  # English
    EN_US = "en-US"  # English (USA)
    EN_AU = "en-AU"  # English (Australia)
    EN_GB = "en-GB"  # English (Great Britain)
    EN_NZ = "en-NZ"  # English (New Zealand)
    EN_IN = "en-IN"  # English (India)
    ET = "et"  # Estonian
    FI = "fi"  # Finnish
    NL_BE = "nl-BE"  # Flemmish
    FR = "fr"  # French
    FR_CA = "fr-CA"  # French (Canada)
    DE = "de"  # German
    DE_CH = "de-CH"  # German (Switzerland)
    EL = "el"  # Greek
    HI = "hi"  # Hindi
    HU = "hu"  # Hungarian
    ID = "id"  # Indonesian
    IT = "it"  # Italian
    JA = "ja"  # Japanese
    KO = "ko"  # Korean
    LV = "lv"  # Latvian
    LT = "lt"  # Lithuanian
    MS = "ms"  # Malay
    NO = "no"  # Norwegian
    PL = "pl"  # Polish
    PT = "pt"  # Portuguese
    PT_BR = "pt-BR"  # Portuguese (Brazil)
    RO = "ro"  # Romanian
    RU = "ru"  # Russian
    SK = "sk"  # Slovak
    ES = "es"  # Spanish
    SV = "sv"  # Swedish
    TH = "th"  # Thai
    TR = "tr"  # Turkish
    UK = "uk"  # Ukrainian
    VI = "vi"  # Vietnamese


TO_LLM_LANGUAGE = {
    "bg": "Bulgarian",
    "ca": "Catalan",
    "": "Chinese simplified(简体中文)",
    "zh": "Chinese simplified(简体中文)",
    "zh-CN": "Chinese simplified(简体中文)",
    "zh-TW": "Chinese traditional(繁体中文)",
    "zn": "Chinese and English",
    "cs": "Czech",
    "da": "Danish",
    "nl": "Dutch",
    "en": "English",
    "en-US": "English (USA)",
    "en-AU": "English (Australia)",
    "en-GB": "English (Great Britain)",
    "en-NZ": "English (New Zealand)",
    "en-IN": "English (India)",
    "et": "Estonian",
    "fi": "Finnish",
    "nl-BE": "Flemmish",
    "fr": "French",
    "fr-CA": "French (Canada)",
    "de": "German",
    "de-CH": "German (Switzerland)",
    "el": "Greek",
    "hi": "Hindi",
    "hu": "Hungarian",
    "id": "Indonesian",
    "it": "Italian",
    "ja": "Japanese",
    "ko": "Korean",
    "lv": "Latvian",
    "lt": "Lithuanian",
    "ms": "Malay",
    "no": "Norwegian",
    "pl": "Polish",
    "pt": "Portuguese",
    "pt-BR": "Portuguese (Brazil)",
    "ro": "Romanian",
    "ru": "Russian",
    "sk": "Slovak",
    "es": "Spanish",
    "sv": "Swedish",
    "th": "Thai",
    "tr": "Turkish",
    "uk": "Ukrainian",
    "vi": "Vietnamese",
}


WHISPER_LANGUAGES = {
    "en": "english",
    "zh": "chinese",
    "de": "german",
    "es": "spanish",
    "ru": "russian",
    "ko": "korean",
    "fr": "french",
    "ja": "japanese",
    "pt": "portuguese",
    "tr": "turkish",
    "pl": "polish",
    "ca": "catalan",
    "nl": "dutch",
    "ar": "arabic",
    "sv": "swedish",
    "it": "italian",
    "id": "indonesian",
    "hi": "hindi",
    "fi": "finnish",
    "vi": "vietnamese",
    "he": "hebrew",
    "uk": "ukrainian",
    "el": "greek",
    "ms": "malay",
    "cs": "czech",
    "ro": "romanian",
    "da": "danish",
    "hu": "hungarian",
    "ta": "tamil",
    "no": "norwegian",
    "th": "thai",
    "ur": "urdu",
    "hr": "croatian",
    "bg": "bulgarian",
    "lt": "lithuanian",
    "la": "latin",
    "mi": "maori",
    "ml": "malayalam",
    "cy": "welsh",
    "sk": "slovak",
    "te": "telugu",
    "fa": "persian",
    "lv": "latvian",
    "bn": "bengali",
    "sr": "serbian",
    "az": "azerbaijani",
    "sl": "slovenian",
    "kn": "kannada",
    "et": "estonian",
    "mk": "macedonian",
    "br": "breton",
    "eu": "basque",
    "is": "icelandic",
    "hy": "armenian",
    "ne": "nepali",
    "mn": "mongolian",
    "bs": "bosnian",
    "kk": "kazakh",
    "sq": "albanian",
    "sw": "swahili",
    "gl": "galician",
    "mr": "marathi",
    "pa": "punjabi",
    "si": "sinhala",
    "km": "khmer",
    "sn": "shona",
    "yo": "yoruba",
    "so": "somali",
    "af": "afrikaans",
    "oc": "occitan",
    "ka": "georgian",
    "be": "belarusian",
    "tg": "tajik",
    "sd": "sindhi",
    "gu": "gujarati",
    "am": "amharic",
    "yi": "yiddish",
    "lo": "lao",
    "uz": "uzbek",
    "fo": "faroese",
    "ht": "haitian creole",
    "ps": "pashto",
    "tk": "turkmen",
    "nn": "nynorsk",
    "mt": "maltese",
    "sa": "sanskrit",
    "lb": "luxembourgish",
    "my": "myanmar",
    "bo": "tibetan",
    "tl": "tagalog",
    "mg": "malagasy",
    "as": "assamese",
    "tt": "tatar",
    "haw": "hawaiian",
    "ln": "lingala",
    "ha": "hausa",
    "ba": "bashkir",
    "jw": "javanese",
    "su": "sundanese",
    "yue": "cantonese",
}


TRANSLATE_LANGUAGE = {
    "ar": "Arabic",
    "cs": "Czech",
    "da": "Danish",
    "de": "German",
    "en": "English",
    "es": "Spanish",
    "fi": "Finnish",
    "fr": "French",
    "hr": "Croatian",
    "hu": "Hungarian",
    "id": "Indonesian",
    "it": "Italian",
    "ja": "Japanese",
    "ko": "Korean",
    "ms": "Malay",
    "nl": "Dutch",
    "nb": "Norwegian Bokmal",
    "no": "Norwegian",
    "pl": "Polish",
    "pt": "Portuguese",
    "ro": "Romanian",
    "ru": "Russian",
    "sv": "Swedish",
    "th": "Thai",
    "tr": "Turkish",
    "uk": "Ukrainian",
    "vi": "Vietnamese",
    "zh": "Chinese",
}



import uuid
import asyncio
import logging
from typing import List, Optional

from apipeline.processors.async_frame_processor import FrameDirection
from apipeline.frames import (
    CancelFrame,
    StartFrame,
    EndFrame,
    Frame,
    TextFrame,
    StartInterruptionFrame,
    StopInterruptionFrame,
    InterruptionFrame,
    InterruptionTaskFrame,
)
from google.adk.artifacts import BaseArtifactService
from google.adk.memory.base_memory_service import BaseMemoryService
from google.adk.sessions.base_session_service import BaseSessionService
from google.genai import types

from src.processors.session_processor import SessionProcessor
from src.services.a2a_live.adk_host_agent_manager import ADKHostLiveAgentManager
from src.services.help.httpx import HTTPXClientWrapper
from src.common.session import Session
from src.common.utils.time import time_now_iso8601
from src.types.frames import (
    VisionImageVoiceRawFrame,
    AudioRawFrame,
    ImageRawFrame,
    VisionImageRawFrame,
    TranscriptionFrame,
    UserImageRawFrame,
)
from src.services.a2a_live.types import LiveMultiModalInputMessage
from src.common.utils.img_utils import process_image_to_base64


class A2ALiveProcessor(SessionProcessor):
    def __init__(
        self,
        app_name: str,
        host_agent_name: str = "",
        voice_name: str = "",
        api_key: str = "",
        mode: str = "supervisor",
        model: str = "",
        system_prompt: str = "",
        http_timeout: float = 30.0,
        agent_urls: Optional[list[str]] = [],  # agent http url
        max_workers: int = 0,
        session: Session | None = None,
        session_service: BaseSessionService | None = None,
        memory_service: BaseMemoryService | None = None,
        artifact_service: BaseArtifactService | None = None,
        **kwargs,
    ):
        super().__init__(session=session, **kwargs)
        self.http_client_wrapper = HTTPXClientWrapper()
        self.http_client_wrapper.start(timeout=http_timeout)
        http_client = self.http_client_wrapper()
        self.manager = ADKHostLiveAgentManager(
            http_client,
            app_name,
            host_agent_name=host_agent_name,
            api_key=api_key,
            mode=mode,
            model=model,
            system_prompt=system_prompt,
            remote_agent_addresses=agent_urls,
            session_service=session_service,
            memory_service=memory_service,
            artifact_service=artifact_service,
            voice_name=voice_name,
        )
        self.manager.initialize_host()
        self.push_task: Optional[asyncio.Task] = None
        self.conversation = None

    @property
    def host_agent_manager(self):
        return self.manager

    def set_user_id(self, user_id: str):
        self.manager.set_user_id(user_id)

    async def create_conversation(self):
        self.conversation = await self.manager.create_conversation()

    async def register_agents(self, urls: List[str]):
        for url in urls:
            await self.manager.register_agent(url)

    async def create_push_task(self):
        if self.push_task is not None:
            self.push_task.cancel()
            try:
                await self.push_task
            except asyncio.CancelledError:
                logging.info(f"{self.name} push_task cancelled.")
            self.push_task = None

        if self.push_task is None:
            self.push_task = self.get_event_loop().create_task(self._push_frames())
            logging.info(f"{self.name} push_task created.")

    async def start(self, frame: StartFrame):
        await self.create_conversation()
        await self.create_push_task()

        if self.http_client_wrapper is None:
            self.http_client_wrapper = HTTPXClientWrapper()
            self.http_client_wrapper.start()
            self.manager.set_client(self.http_client_wrapper())

        logging.info(f"{self.name} Conversation started")

    async def stop(self, frame: EndFrame):
        if self.http_client_wrapper:
            await self.http_client_wrapper.stop()
            self.http_client_wrapper = None

        logging.info(f"{self.name} Conversation end")

    async def cancel(self, frame: CancelFrame):
        self.push_task.cancel()
        try:
            await self.push_task
        except asyncio.CancelledError:
            logging.info(f"{self.name} push_task cancelled.")
        if self.http_client_wrapper:
            await self.http_client_wrapper.stop()
            self.http_client_wrapper = None
        logging.info(f"{self.name} Conversation cancelled")

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        """Process incoming frames, intercept context frames for memory integration.

        Args:
            frame: The incoming frame to process.
            direction: The direction of frame flow in the pipeline.
        """
        await super().process_frame(frame, direction)

        if isinstance(frame, StartFrame):
            await self.start(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, EndFrame):
            await self.stop(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, CancelFrame):
            await self.cancel(frame)
            await self.push_frame(frame, direction)
        elif isinstance(frame, VisionImageRawFrame):
            await self.manager.send_live_message(
                LiveMultiModalInputMessage(
                    context_id=self.conversation.conversation_id,
                    message_id=str(uuid.uuid4()),
                    kind="text_images",
                    text_content=types.Content(role="user", parts=[types.Part(text=frame.text)]),
                    image_blob_list=[
                        types.Blob(
                            data=process_image_to_base64(frame),
                            mime_type="image/jpeg",
                        )
                    ],
                )
            )
        elif isinstance(frame, VisionImageVoiceRawFrame):
            await self.manager.send_live_message(
                LiveMultiModalInputMessage(
                    context_id=self.conversation.conversation_id,
                    message_id=str(uuid.uuid4()),
                    kind="audio_images",
                    audio_blob=types.Blob(
                        data=frame.audio.audio,
                        mime_type=f"audio/pcm;rate={frame.audio.sample_rate}",
                    ),
                    image_blob_list=[
                        types.Blob(
                            data=process_image_to_base64(img),
                            mime_type="image/jpeg",
                        )
                        for img in frame.images
                    ],
                )
            )
        elif isinstance(frame, TextFrame):
            await self.manager.send_live_message(
                LiveMultiModalInputMessage(
                    context_id=self.conversation.conversation_id,
                    message_id=str(uuid.uuid4()),
                    kind="text",
                    text_content=types.Content(role="user", parts=[types.Part(text=frame.text)]),
                )
            )
        elif isinstance(frame, AudioRawFrame):
            await self.manager.send_live_message(
                LiveMultiModalInputMessage(
                    context_id=self.conversation.conversation_id,
                    message_id=str(uuid.uuid4()),
                    kind="audio",
                    audio_blob=types.Blob(
                        data=frame.audio,
                        mime_type=f"audio/pcm;rate={frame.sample_rate}",
                    ),
                )
            )
        elif isinstance(frame, ImageRawFrame):
            await self.manager.send_live_message(
                LiveMultiModalInputMessage(
                    context_id=self.conversation.conversation_id,
                    message_id=str(uuid.uuid4()),
                    kind="images",
                    image_blob_list=[
                        types.Blob(
                            data=process_image_to_base64(frame),
                            mime_type="image/jpeg",
                        )
                    ],
                ),
            )
        else:
            await self.queue_frame(frame, direction)

    async def _push_frames(self):
        try:
            async for msg in self.manager.recieve_message():
                if self._allow_interruptions and msg.kind == "interrupted" and msg.interrupted:
                    logging.info(f"{self.name} Interrupting")
                    await self._handle_interruptions(InterruptionFrame())
                    logging.info(f"{self.name} Interrupted")
                if msg.kind == "transcription":
                    await self.queue_frame(
                        TranscriptionFrame(
                            text=msg.text_content.parts[0].text,
                            user_id=self.session.ctx.client_id,
                            timestamp=time_now_iso8601(),
                        )
                    )
                if msg.kind == "text":
                    await self.queue_frame(TextFrame(text=msg.text_content.parts[0].text))
                if msg.kind == "audio":
                    rate = int(msg.audio_blob.mime_type.split("=")[1])
                    await self.queue_frame(
                        AudioRawFrame(audio=msg.audio_blob.data, sample_rate=rate)
                    )
                if (
                    msg.kind == "images"
                ):  # gemini live 2.5 no image , wait gemini live 3.0 (maybe gen image)
                    # TODO: add image frame support
                    pass
        except Exception as e:
            logging.exception(f"{self.name} Error in _push_frames: {e}")



from .qwen2_mtp_v4_48_3 import Qwen2MTPConfig, Qwen2MTPForCausalLM
from .qwen2_mtp_sensevoice_v4_48_3 import Qwen2MTPSenseVoiceConfig, Qwen2MTPSenseVoiceForCausalLM


# from .qwen2_v4_48_3 import Qwen2Config, Qwen2ForCausalLM

